"""GitHub OAuth provider implementation."""

from __future__ import annotations

from urllib.parse import urlencode

import httpx

from derp.auth.exceptions import OAuthProviderError
from derp.auth.providers.base import BaseOAuthProvider, OAuthTokens, OAuthUserInfo
from derp.config import GitHubOAuthConfig


class GitHubProvider(BaseOAuthProvider[GitHubOAuthConfig]):
    """GitHub OAuth provider."""

    provider_name = "github"

    # GitHub OAuth endpoints
    AUTHORIZATION_URL = "https://github.com/login/oauth/authorize"
    TOKEN_URL = "https://github.com/login/oauth/access_token"
    USERINFO_URL = "https://api.github.com/user"
    EMAILS_URL = "https://api.github.com/user/emails"

    def get_authorization_url(
        self,
        state: str,
        scopes: list[str] | None = None,
        redirect_uri: str | None = None,
    ) -> str:
        """Generate GitHub OAuth authorization URL."""
        params = {
            "client_id": self._config.client_id,
            "redirect_uri": redirect_uri or self._config.redirect_uri,
            "state": state,
            "scope": " ".join(scopes or self._config.scopes),
        }
        return f"{self.AUTHORIZATION_URL}?{urlencode(params)}"

    async def exchange_code(
        self,
        code: str,
        redirect_uri: str | None = None,
    ) -> OAuthTokens:
        """Exchange authorization code for GitHub tokens."""
        data = {
            "client_id": self._config.client_id,
            "client_secret": self._config.client_secret,
            "code": code,
            "redirect_uri": redirect_uri or self._config.redirect_uri,
        }

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.TOKEN_URL,
                data=data,
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
            )

            if response.status_code != 200:
                raise OAuthProviderError(f"Failed to exchange code: {response.text}")

            token_data = response.json()

            if "error" in token_data:
                message = token_data.get("error_description", token_data["error"])
                raise OAuthProviderError(f"Failed to exchange code: {message}")

        return OAuthTokens(
            access_token=token_data["access_token"],
            token_type=token_data.get("token_type", "bearer"),
            scope=token_data.get("scope"),
        )

    async def get_user_info(self, access_token: str) -> OAuthUserInfo:
        """Get user info from GitHub."""
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        }

        async with httpx.AsyncClient() as client:
            # Get basic user info
            response = await client.get(self.USERINFO_URL, headers=headers)

            if response.status_code != 200:
                raise OAuthProviderError(f"Failed to get user info: {response.text}")

            user_data = response.json()

            # Get user emails (GitHub may not include email in user response)
            email = user_data.get("email")
            email_verified = False

            if not email:
                email_response = await client.get(self.EMAILS_URL, headers=headers)
                if email_response.status_code == 200:
                    emails = email_response.json()
                    # Find primary verified email
                    for email_entry in emails:
                        if email_entry.get("primary") and email_entry.get("verified"):
                            email = email_entry["email"]
                            email_verified = True
                            break
                    # Fallback to any verified email
                    if not email:
                        for email_entry in emails:
                            if email_entry.get("verified"):
                                email = email_entry["email"]
                                email_verified = True
                                break
                    # Last resort: any email
                    if not email and emails:
                        email = emails[0]["email"]

            if not email:
                raise OAuthProviderError(
                    "Could not retrieve email from GitHub. "
                    "Ensure the user:email scope is granted."
                )

        return OAuthUserInfo(
            id=str(user_data["id"]),
            email=email,
            email_verified=email_verified,
            name=user_data.get("name") or user_data.get("login"),
            picture=user_data.get("avatar_url"),
            raw_data=user_data,
        )



"""KV client interfaces and backends."""

from derp.config import ValkeyMode
from derp.kv.base import KVClient
from derp.kv.valkey import ValkeyClient, ValkeyConfig

__all__ = [
    "KVClient",
    "ValkeyClient",
    "ValkeyConfig",
    "ValkeyMode",
]



"""Base interface for queue clients."""

from __future__ import annotations

import abc
import dataclasses
import enum
from datetime import timedelta
from typing import Any


class TaskState(enum.StrEnum):
    """State of a queued task."""

    PENDING = "pending"
    STARTED = "started"
    SUCCESS = "success"
    FAILURE = "failure"
    REVOKED = "revoked"
    UNKNOWN = "unknown"


@dataclasses.dataclass(slots=True)
class TaskStatus:
    """Status of a queued task."""

    task_id: str
    state: TaskState
    result: Any | None = None
    error: str | None = None


class ScheduleType(enum.StrEnum):
    """Type of recurring schedule."""

    CRON = "cron"
    INTERVAL = "interval"


@dataclasses.dataclass(slots=True, frozen=True)
class Schedule:
    """A recurring task schedule."""

    name: str
    task: str
    type: ScheduleType
    cron: str | None = None
    interval: timedelta | None = None
    payload: dict[str, Any] | None = None
    queue: str | None = None
    path: str | None = None


class QueueClient(abc.ABC):
    """Async producer-side queue client."""

    supports_result: bool
    supports_revoke: bool
    supports_delay: bool

    @abc.abstractmethod
    async def connect(self) -> None:
        """Connect to the queue backend."""

    @abc.abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from the queue backend."""

    @abc.abstractmethod
    async def enqueue(
        self,
        task_name: str,
        payload: dict[str, Any] | None = None,
        *,
        queue: str | None = None,
        delay: timedelta | None = None,
    ) -> str:
        """Enqueue a task. Returns a task ID."""

    @abc.abstractmethod
    async def get_status(self, task_id: str) -> TaskStatus:
        """Get the status of a task by ID."""



"""Module loading utilities for Derp ORM."""

from __future__ import annotations

import importlib
import sys
from pathlib import Path

from derp.orm.table import Table


def _load_tables_from_file(path: Path) -> list[type[Table]]:
    """Load Table subclasses from a single Python file."""

    try:
        # Convert file path to dotted module name relative to CWD
        resolved = path.resolve()
        cwd = Path.cwd()
        try:
            relative = resolved.relative_to(cwd)
        except ValueError:
            # Absolute path outside CWD — use just the stem and add parent to sys.path
            relative = Path(resolved.stem + ".py")
            parent = str(resolved.parent)
            if parent not in sys.path:
                sys.path.insert(0, parent)

        module_name = ".".join(relative.with_suffix("").parts)

        # Ensure CWD is on sys.path so relative module names resolve
        cwd_str = str(cwd)
        if cwd_str not in sys.path:
            sys.path.insert(0, cwd_str)

        module = importlib.import_module(module_name)
    except ImportError as e:
        raise ImportError(f"Failed to import module at {path.name}.") from e

    # Find all Table subclasses
    tables: list[type[Table]] = []
    for name in dir(module):
        obj = getattr(module, name)
        if (
            isinstance(obj, type)
            and issubclass(obj, Table)
            and obj is not Table
            and hasattr(obj, "__columns__")
            and getattr(obj, "__explicit_table__", False)
        ):
            tables.append(obj)

    return tables


def load_tables(module_path: str) -> list[type[Table]]:
    """Load Table subclasses from a Python module, directory, or glob pattern.

    Args:
        module_path: One of:
            - Path to a Python file (e.g., "src/myapp/schema.py")
            - Path to a directory (loads all .py files recursively)
            - Glob pattern (e.g., "src/**/models.py", "src/**/*.py")

    Returns:
        List of Table subclasses found in matching modules
    """
    path = Path(module_path)
    tables: list[type[Table]] = []
    seen_tables: set[int] = set()  # Track by id to avoid duplicates

    # Check if it's a glob pattern
    if "*" in module_path or "?" in module_path or "[" in module_path:
        # Find the base directory (everything before the first glob character)
        parts = Path(module_path).parts
        base_parts: list[str] = []
        pattern_parts: list[str] = []
        found_glob = False

        for part in parts:
            if found_glob or "*" in part or "?" in part or "[" in part:
                found_glob = True
                pattern_parts.append(part)
            else:
                base_parts.append(part)

        base_dir = Path(*base_parts) if base_parts else Path(".")
        pattern = str(Path(*pattern_parts)) if pattern_parts else "**/*.py"

        if not base_dir.exists():
            raise FileNotFoundError(f"Base directory not found: {base_dir}")

        files = sorted(base_dir.glob(pattern))

    elif path.is_dir():
        # Directory: load all .py files recursively
        files = sorted(path.rglob("*.py"))

    elif path.is_file():
        # Single file
        files = [path]

    else:
        raise FileNotFoundError(f"Path not found: {module_path}")

    # Load tables from each file
    for file_path in files:
        if file_path.name.startswith("_"):
            continue  # Skip __init__.py, __pycache__, etc.

        try:
            for table in _load_tables_from_file(file_path):
                # Use qualified name to deduplicate
                key: int = id(table)
                if key not in seen_tables:
                    seen_tables.add(key)
                    tables.append(table)
        except Exception:
            # Skip files that fail to import (might not be valid modules)
            continue

    return tables


def _deduplicate_tables(tables: list[type[Table]]) -> list[type[Table]]:
    """Keep only the youngest table in each inheritance chain.

    If a table has a descendant in the list, drop it (the descendant
    inherits all its columns). If two tables share a common explicit-table
    ancestor but neither descends from the other, raise ValueError.
    """
    # Find tables that are ancestors of another table in the list
    to_remove: set[type[Table]] = set()
    for t in tables:
        for other in tables:
            if other is t:
                continue
            # If t is a proper ancestor of other, t is redundant
            if issubclass(other, t) and t is not Table:
                to_remove.add(t)

    result = [t for t in tables if t not in to_remove]

    # Check for branches: two result tables sharing a common explicit-table
    # ancestor (whether or not that ancestor was in the input list)
    for i, a in enumerate(result):
        for b in result[i + 1 :]:
            # Walk a's MRO looking for a shared explicit-table ancestor
            for base in type.mro(a):
                if (
                    base is a
                    or base is b
                    or base is Table
                    or not isinstance(base, type)
                ):
                    continue
                if (
                    getattr(base, "__explicit_table__", False)
                    and issubclass(a, base)
                    and issubclass(b, base)
                ):
                    raise ValueError(
                        f"Ambiguous table inheritance: {a.__name__} and "
                        f"{b.__name__} both extend {base.__name__}. "
                        f"Only one subclass per table hierarchy is allowed."
                    )

    return result


def discover_tables(
    schema_path: str,
    *,
    include_auth: bool = False,
) -> list[type[Table]]:
    """Load user tables and optionally inject framework tables, then dedup.

    Args:
        schema_path: Path/glob/directory for user schema files.
        include_auth: If True, inject AuthUser and AuthSession when no
            subclass is already present in the loaded tables.

    Returns:
        Deduplicated list of Table subclasses.
    """
    tables = load_tables(schema_path)

    if include_auth:
        from derp.auth.models import (
            AuthOrganization,
            AuthOrgMember,
            AuthSession,
            AuthUser,
        )

        for auth_table in (AuthUser, AuthSession, AuthOrganization, AuthOrgMember):
            if not any(issubclass(t, auth_table) for t in tables):
                tables.append(auth_table)

    return _deduplicate_tables(tables)



"""Index operation convertors."""

from __future__ import annotations

from derp.orm.migrations.convertors.base import (
    ConvertorRegistry,
    StatementConvertor,
    quote_identifier,
    quote_schema_table,
)
from derp.orm.migrations.statements.types import (
    CreateIndexStatement,
    DropIndexStatement,
)


class CreateIndexConvertor(StatementConvertor[CreateIndexStatement]):
    """Convert CREATE INDEX statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "create_index"

    def convert(self, statement: CreateIndexStatement) -> str:
        parts = ["CREATE"]

        if statement.unique:
            parts.append("UNIQUE")

        parts.append("INDEX")

        if statement.concurrently:
            parts.append("CONCURRENTLY")

        parts.append(quote_identifier(statement.name))

        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        parts.append(f"ON {table_ref}")

        # Index method
        if statement.method and statement.method.lower() != "btree":
            parts.append(f"USING {statement.method.upper()}")

        # Columns
        cols = ", ".join(quote_identifier(c) for c in statement.columns)
        parts.append(f"({cols})")

        # INCLUDE columns
        if statement.include:
            include_cols = ", ".join(quote_identifier(c) for c in statement.include)
            parts.append(f"INCLUDE ({include_cols})")

        # NULLS NOT DISTINCT (PostgreSQL 15+)
        if statement.nulls_not_distinct:
            parts.append("NULLS NOT DISTINCT")

        # WHERE clause for partial indexes
        if statement.where:
            parts.append(f"WHERE {statement.where}")

        return " ".join(parts) + ";"


class DropIndexConvertor(StatementConvertor[DropIndexStatement]):
    """Convert DROP INDEX statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "drop_index"

    def convert(self, statement: DropIndexStatement) -> str:
        parts = ["DROP INDEX"]

        if statement.concurrently:
            parts.append("CONCURRENTLY")

        parts.append("IF EXISTS")

        # Index with schema
        if statement.schema_name and statement.schema_name != "public":
            parts.append(
                f"{quote_identifier(statement.schema_name)}.{quote_identifier(statement.name)}"
            )
        else:
            parts.append(quote_identifier(statement.name))

        return " ".join(parts) + ";"


# Register convertors
ConvertorRegistry.register(CreateIndexConvertor())
ConvertorRegistry.register(DropIndexConvertor())



"""JWT token creation and validation."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta
from typing import Any

import jwt

from derp.config import JWTConfig


@dataclass(kw_only=True)
class TokenPayload:
    """Payload data from a decoded JWT token."""

    sub: str  # User ID
    session_id: str
    exp: datetime
    iat: datetime
    iss: str | None = None
    aud: str | None = None
    extra: dict[str, Any] | None = None


@dataclass(kw_only=True)
class TokenPair:
    """Access and refresh token pair."""

    access_token: str
    refresh_token: str
    token_type: str = "bearer"
    expires_in: int = 0  # Access token expiry in seconds
    expires_at: datetime


def create_access_token(
    config: JWTConfig,
    user_id: str | uuid.UUID,
    session_id: str | uuid.UUID,
    extra_claims: dict[str, Any] | None = None,
) -> str:
    """Create a short-lived JWT access token.

    Args:
        user_id: The user's unique identifier
        session_id: The session's unique identifier
        extra_claims: Additional claims to include in the token

    Returns:
        Encoded JWT token string
    """
    now = datetime.now(UTC)
    expires = now + timedelta(minutes=config.access_token_expire_minutes)

    payload: dict[str, Any] = {
        "sub": str(user_id),
        "session_id": str(session_id),
        "iat": now,
        "exp": expires,
    }

    if config.issuer:
        payload["iss"] = config.issuer

    if config.audience:
        payload["aud"] = config.audience

    if extra_claims:
        reserved = {"sub", "session_id", "iat", "exp", "iss", "aud"}
        conflicts = reserved & extra_claims.keys()
        if conflicts:
            raise ValueError(
                "extra_claims cannot override reserved JWT claims: "
                f"{', '.join(sorted(conflicts))}"
            )
        payload.update(extra_claims)

    return jwt.encode(
        payload,
        config.secret,
        algorithm=config.algorithm,
    )


def decode_token(config: JWTConfig, token: str) -> TokenPayload | None:
    """Decode and validate a JWT token.

    Args:
        token: The JWT token string

    Returns:
        TokenPayload with decoded data, or None if the token is
        expired or invalid.
    """
    try:
        payload = jwt.decode(
            token,
            config.secret,
            algorithms=[config.algorithm],
            audience=config.audience,
            issuer=config.issuer,
            options={"require": ["aud"]} if config.audience else None,
        )
    except jwt.InvalidTokenError:
        return None

    # Extract known fields
    sub = payload.get("sub")
    session_id = payload.get("session_id")
    exp = payload.get("exp")
    iat = payload.get("iat")

    if not sub or not session_id:
        return None

    # Convert timestamps
    exp_dt = datetime.fromtimestamp(exp, tz=UTC) if exp else datetime.now(UTC)
    iat_dt = datetime.fromtimestamp(iat, tz=UTC) if iat else datetime.now(UTC)

    # Extract extra claims
    known_keys = {"sub", "session_id", "exp", "iat", "iss", "aud"}
    extra = {k: v for k, v in payload.items() if k not in known_keys}
    iss = payload.get("iss")
    aud = payload.get("aud")

    return TokenPayload(
        sub=sub,
        session_id=session_id,
        exp=exp_dt,
        iat=iat_dt,
        iss=iss,
        aud=aud,
        extra=extra if extra else None,
    )


def create_token_pair(
    config: JWTConfig,
    user_id: str | uuid.UUID,
    session_id: str | uuid.UUID,
    refresh_token: str,
    extra_claims: dict[str, Any] | None = None,
) -> TokenPair:
    """Create a token pair with access token and refresh token.

    Args:
        user_id: The user's unique identifier
        session_id: The session's unique identifier
        refresh_token: The refresh token string (created separately)
        extra_claims: Additional claims for the access token

    Returns:
        TokenPair with both tokens
    """
    access_token = create_access_token(config, user_id, session_id, extra_claims)
    expires_in = config.access_token_expire_minutes * 60
    expires_at = datetime.now(UTC) + timedelta(
        minutes=config.access_token_expire_minutes
    )

    return TokenPair(
        access_token=access_token,
        refresh_token=refresh_token,
        token_type="bearer",
        expires_in=expires_in,
        expires_at=expires_at,
    )



"""WHERE clause expressions and operators for Derp ORM."""

from __future__ import annotations

import abc
import dataclasses
import re
from collections.abc import Sequence
from enum import StrEnum
from typing import Any

from derp.orm.fields import FieldInfo


class LogicalOperator(StrEnum):
    """SQL logical operators."""

    AND = "AND"
    OR = "OR"


class ComparisonOperator(StrEnum):
    """SQL comparison operators."""

    EQ = "="
    NE = "<>"
    GT = ">"
    GTE = ">="
    LT = "<"
    LTE = "<="


@dataclasses.dataclass
class Expression(abc.ABC):
    """Base class for SQL expressions."""

    @abc.abstractmethod
    def to_sql(self, params: list[Any]) -> str:
        """Generate SQL string with parameterized values.

        Args:
            params: List to append parameter values to

        Returns:
            SQL string with $N placeholders
        """

    def __and__(self, other: Expression) -> Expression:
        return LogicalOp(LogicalOperator.AND, (self, other))

    def __or__(self, other: Expression) -> Expression:
        return LogicalOp(LogicalOperator.OR, (self, other))

    def __invert__(self) -> Expression:
        return UnaryOp("NOT", self)

    def __eq__(self, other: Any) -> Any:
        """Equal comparison: field == value."""
        return BinaryOp(self, ComparisonOperator.EQ, to_expr(other))

    def __ne__(self, other: Any) -> Any:
        """Not equal comparison: field != value."""
        return BinaryOp(self, ComparisonOperator.NE, to_expr(other))

    def __lt__(self, other: Any) -> Any:
        """Less than comparison: field < value."""
        return BinaryOp(self, ComparisonOperator.LT, to_expr(other))

    def __le__(self, other: Any) -> Any:
        """Less than or equal comparison: field <= value."""
        return BinaryOp(self, ComparisonOperator.LTE, to_expr(other))

    def __gt__(self, other: Any) -> Any:
        """Greater than comparison: field > value."""
        return BinaryOp(self, ComparisonOperator.GT, to_expr(other))

    def __ge__(self, other: Any) -> Any:
        """Greater than or equal comparison: field >= value."""
        return BinaryOp(self, ComparisonOperator.GTE, to_expr(other))

    def in_(self, values: Sequence[Any] | Any) -> Any:
        """IN clause. Accepts a list of values or a SelectQuery."""
        if hasattr(values, "build"):
            return InSubquery(self, values, negated=False)
        return InList(self, tuple(values), negated=False)

    def not_in(self, values: Sequence[Any] | Any) -> Any:
        """NOT IN clause. Accepts a list of values or a SelectQuery."""
        if hasattr(values, "build"):
            return InSubquery(self, values, negated=True)
        return InList(self, tuple(values), negated=True)

    def like(self, pattern: str) -> Any:
        """LIKE pattern matching."""
        return Like(self, pattern, case_insensitive=False)

    def ilike(self, pattern: str) -> Any:
        """ILIKE pattern matching."""
        return Like(self, pattern, case_insensitive=True)

    def is_null(self) -> Any:
        """IS NULL check."""
        return NullCheck(self, is_null=True)

    def is_not_null(self) -> Any:
        """IS NOT NULL check."""
        return NullCheck(self, is_null=False)

    def between(self, low: Any, high: Any) -> Any:
        """BETWEEN range check."""
        return Between(self, low, high)


@dataclasses.dataclass(eq=False)
class ColumnRef(Expression):
    """Reference to a table column."""

    table_name: str
    column_name: str

    def to_sql(self, params: list[Any]) -> str:
        return f"{self.table_name}.{self.column_name}"


@dataclasses.dataclass
class Literal(Expression):
    """Literal value."""

    value: Any

    def to_sql(self, params: list[Any]) -> str:
        params.append(self.value)
        return f"${len(params)}"


@dataclasses.dataclass
class BinaryOp(Expression):
    """Binary operator expression (e.g., a = b)."""

    left: Expression | FieldInfo | Any
    operator: ComparisonOperator | str
    right: Expression | FieldInfo | Any

    def to_sql(self, params: list[Any]) -> str:
        left_sql = _expr_to_sql(self.left, params)
        right_sql = _expr_to_sql(self.right, params)
        return f"({left_sql} {self.operator} {right_sql})"


@dataclasses.dataclass
class UnaryOp(Expression):
    """Unary operator expression (e.g., NOT a)."""

    operator: str
    operand: Expression

    def to_sql(self, params: list[Any]) -> str:
        operand_sql = _expr_to_sql(self.operand, params)
        return f"({self.operator} {operand_sql})"


@dataclasses.dataclass
class LogicalOp(Expression):
    """Logical combination of expressions (AND/OR)."""

    operator: LogicalOperator
    conditions: tuple[Expression, ...]

    def to_sql(self, params: list[Any]) -> str:
        if not self.conditions:
            return "TRUE" if self.operator == LogicalOperator.AND else "FALSE"
        parts = [_expr_to_sql(c, params) for c in self.conditions]
        return f"({f' {self.operator} '.join(parts)})"


@dataclasses.dataclass
class InList(Expression):
    """IN expression (a IN (1, 2, 3))."""

    column: Expression | FieldInfo
    values: tuple[Any, ...]
    negated: bool = False

    def to_sql(self, params: list[Any]) -> str:
        if not self.values:
            return "FALSE" if not self.negated else "TRUE"
        col_sql = _expr_to_sql(self.column, params)
        placeholders = []
        for v in self.values:
            params.append(v)
            placeholders.append(f"${len(params)}")
        op = "NOT IN" if self.negated else "IN"
        return f"({col_sql} {op} ({', '.join(placeholders)}))"


@dataclasses.dataclass
class InSubquery(Expression):
    """IN (SELECT ...) expression."""

    column: Expression | FieldInfo
    query: Any  # SelectQuery — typed as Any to avoid circular import
    negated: bool = False

    def to_sql(self, params: list[Any]) -> str:
        col_sql = _expr_to_sql(self.column, params)
        sub_sql, sub_params = self.query.build()
        offset = len(params)
        params.extend(sub_params)
        renumbered = _renumber_params(sub_sql, offset)
        op = "NOT IN" if self.negated else "IN"
        return f"({col_sql} {op} ({renumbered}))"


@dataclasses.dataclass
class Between(Expression):
    """BETWEEN expression."""

    column: Expression | FieldInfo
    low: Any
    high: Any

    def to_sql(self, params: list[Any]) -> str:
        col_sql = _expr_to_sql(self.column, params)
        params.append(self.low)
        low_placeholder = f"${len(params)}"
        params.append(self.high)
        high_placeholder = f"${len(params)}"
        return f"({col_sql} BETWEEN {low_placeholder} AND {high_placeholder})"


@dataclasses.dataclass
class NullCheck(Expression):
    """IS NULL / IS NOT NULL expression."""

    column: Expression | FieldInfo
    is_null: bool = True

    def to_sql(self, params: list[Any]) -> str:
        col_sql = _expr_to_sql(self.column, params)
        op = "IS NULL" if self.is_null else "IS NOT NULL"
        return f"({col_sql} {op})"


@dataclasses.dataclass
class Like(Expression):
    """LIKE/ILIKE pattern matching."""

    column: Expression | FieldInfo
    pattern: str
    case_insensitive: bool = False

    def to_sql(self, params: list[Any]) -> str:
        col_sql = _expr_to_sql(self.column, params)
        params.append(self.pattern)
        op = "ILIKE" if self.case_insensitive else "LIKE"
        return f"({col_sql} {op} ${len(params)})"


def _expr_to_sql(expr: Expression | FieldInfo | Any, params: list[Any]) -> str:
    """Convert an expression, FieldInfo, or literal to SQL."""

    if isinstance(expr, Expression):
        return expr.to_sql(params)
    elif isinstance(expr, FieldInfo):
        # FieldInfo stores table and column name
        if expr._table_name and expr._field_name:
            return f"{expr._table_name}.{expr._field_name}"
        elif expr._field_name:
            return expr._field_name
        raise ValueError("FieldInfo missing table/column name metadata")
    else:
        # Literal value
        params.append(expr)
        return f"${len(params)}"


@dataclasses.dataclass
class RawSQL(Expression):
    """Raw SQL fragment with optional parameterized values.

    Use the ``sql()`` factory function to create instances::

        sql("NOW()")
        sql("age > {}", 18)
        sql("age > {} AND name = {}", 18, "Alice")
    """

    template: str
    values: tuple[Any, ...]
    _alias: str | None = dataclasses.field(default=None, repr=False)

    def to_sql(self, params: list[Any]) -> str:
        parts = self.template.split("{}")
        result = parts[0]
        for i, val in enumerate(self.values):
            params.append(val)
            result += f"${len(params)}"
            if i + 1 < len(parts):
                result += parts[i + 1]
        if self._alias is not None:
            result += f" AS {self._alias}"
        return result

    def as_(self, alias: str) -> RawSQL:
        """Return a copy with an AS alias."""
        return RawSQL(self.template, self.values, _alias=alias)


def sql(template: str, *values: Any) -> RawSQL:
    """Create a raw SQL expression fragment.

    Use ``{}`` as placeholder for parameterized values::

        sql("NOW()")
        sql("age > {}", 18)
        sql("CONCAT({}, name)", "Dr. ")
    """
    return RawSQL(template, values)


@dataclasses.dataclass
class AggregateFunc(Expression):
    """SQL aggregate function expression (COUNT, SUM, AVG, MIN, MAX)."""

    func: str
    arg: Expression | FieldInfo
    _alias: str | None = dataclasses.field(default=None, repr=False)

    def to_sql(self, params: list[Any]) -> str:
        arg_sql = _expr_to_sql(self.arg, params)
        result = f"{self.func}({arg_sql})"
        if self._alias is not None:
            result += f" AS {self._alias}"
        return result

    def as_(self, alias: str) -> AggregateFunc:
        """Return a copy with an AS alias."""
        return AggregateFunc(self.func, self.arg, _alias=alias)


@dataclasses.dataclass
class CaseExpression(Expression):
    """SQL CASE expression.

    Simple CASE: ``CASE operand WHEN val THEN result ... END``
    """

    operand: Expression | FieldInfo
    whens: list[tuple[Any, Any]]
    else_value: Any | None = dataclasses.field(default=None)
    _alias: str | None = dataclasses.field(default=None, repr=False)

    def to_sql(self, params: list[Any]) -> str:
        operand_sql = _expr_to_sql(self.operand, params)
        result = f"CASE {operand_sql}"
        for cond, val in self.whens:
            params.append(cond)
            cond_ph = f"${len(params)}"
            params.append(val)
            val_ph = f"${len(params)}"
            result += f" WHEN {cond_ph} THEN {val_ph}"
        if self.else_value is not None:
            params.append(self.else_value)
            result += f" ELSE ${len(params)}"
        result += " END"
        if self._alias is not None:
            result += f" AS {self._alias}"
        return result

    def as_(self, alias: str) -> CaseExpression:
        """Return a copy with an AS alias."""
        return CaseExpression(self.operand, self.whens, self.else_value, _alias=alias)


_PARAM_RE = re.compile(r"\$(\d+)")


def _renumber_params(sql_str: str, offset: int) -> str:
    """Shift all ``$N`` placeholders in *sql_str* by *offset*."""
    if offset == 0:
        return sql_str
    return _PARAM_RE.sub(lambda m: f"${int(m.group(1)) + offset}", sql_str)


@dataclasses.dataclass
class SubqueryExpr(Expression):
    """A SELECT query wrapped as an expression (for use in WHERE/SELECT/FROM)."""

    query: Any  # SelectQuery — typed as Any to avoid circular import
    _alias: str | None = dataclasses.field(default=None, repr=False)

    def to_sql(self, params: list[Any]) -> str:
        sub_sql, sub_params = self.query.build()
        offset = len(params)
        params.extend(sub_params)
        renumbered = _renumber_params(sub_sql, offset)
        result = f"({renumbered})"
        if self._alias is not None:
            result += f" AS {self._alias}"
        return result

    def as_(self, alias: str) -> SubqueryExpr:
        """Return a copy with an AS alias."""
        return SubqueryExpr(self.query, _alias=alias)


@dataclasses.dataclass
class ExistsExpr(Expression):
    """EXISTS (SELECT ...) expression."""

    subquery: SubqueryExpr

    def to_sql(self, params: list[Any]) -> str:
        sub_sql = self.subquery.to_sql(params)
        return f"EXISTS {sub_sql}"


def to_expr(value: Expression | FieldInfo | Any) -> Expression | FieldInfo:
    """Ensure value is an expression or FieldInfo."""
    if isinstance(value, Expression | FieldInfo):
        return value
    return Literal(value)



"""Pydantic models for database schema snapshots.

These models represent the complete state of a database schema at a point in time,
matching Drizzle's snapshot format for PostgreSQL.
"""

from __future__ import annotations

from collections.abc import Sequence
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field, PrivateAttr


class SnapshotVersion(StrEnum):
    """Snapshot format version."""

    V1 = "1"


class ForeignKeyAction(StrEnum):
    """Foreign key referential actions."""

    CASCADE = "cascade"
    SET_NULL = "set null"
    SET_DEFAULT = "set default"
    RESTRICT = "restrict"
    NO_ACTION = "no action"


class IndexMethod(StrEnum):
    """PostgreSQL index methods."""

    BTREE = "btree"
    HASH = "hash"
    GIN = "gin"
    GIST = "gist"
    SPGIST = "spgist"
    BRIN = "brin"


class IdentityGeneration(StrEnum):
    """Identity column generation type."""

    ALWAYS = "always"
    BY_DEFAULT = "by default"


class IdentityConfig(BaseModel):
    """IDENTITY column configuration."""

    generation: IdentityGeneration = IdentityGeneration.BY_DEFAULT
    start: int = 1
    increment: int = 1
    min_value: int | None = None
    max_value: int | None = None
    cycle: bool = False


class ColumnSnapshot(BaseModel):
    """Serialized column definition."""

    name: str
    type: str  # e.g., "varchar(255)", "serial", "integer"
    primary_key: bool = False
    not_null: bool = True
    unique: bool = False
    default: str | None = None  # SQL expression or literal
    generated: str | None = None  # Generated column expression (STORED)
    identity: IdentityConfig | None = None
    array_dimensions: int = 0  # For array types

    class ConfigDict:
        use_enum_values = True


class ForeignKeySnapshot(BaseModel):
    """Serialized foreign key constraint."""

    name: str
    columns: list[str]
    references_schema: str = "public"
    references_table: str
    references_columns: list[str]
    on_delete: ForeignKeyAction | None = None
    on_update: ForeignKeyAction | None = None
    deferrable: bool = False
    initially_deferred: bool = False

    class ConfigDict:
        use_enum_values = True


class IndexSnapshot(BaseModel):
    """Serialized index definition."""

    name: str
    columns: list[str]
    unique: bool = False
    where: str | None = None  # Partial index condition
    method: IndexMethod = IndexMethod.BTREE
    concurrently: bool = False
    nulls_not_distinct: bool = False  # PostgreSQL 15+
    include: list[str] = Field(default_factory=list)  # INCLUDE columns
    with_options: dict[str, str] = Field(default_factory=dict)  # WITH (option=value)

    class ConfigDict:
        use_enum_values = True


class UniqueConstraintSnapshot(BaseModel):
    """Serialized unique constraint."""

    name: str
    columns: list[str]
    nulls_not_distinct: bool = False  # PostgreSQL 15+


class CheckConstraintSnapshot(BaseModel):
    """Serialized check constraint."""

    name: str
    expression: str


class PrimaryKeySnapshot(BaseModel):
    """Serialized primary key constraint."""

    name: str | None = None
    columns: list[str]


class TableSnapshot(BaseModel):
    """Serialized table definition."""

    name: str
    schema_name: str = Field(default="public", alias="schema")
    columns: dict[str, ColumnSnapshot] = Field(default_factory=dict)
    primary_key: PrimaryKeySnapshot | None = None
    foreign_keys: dict[str, ForeignKeySnapshot] = Field(default_factory=dict)
    indexes: dict[str, IndexSnapshot] = Field(default_factory=dict)
    unique_constraints: dict[str, UniqueConstraintSnapshot] = Field(
        default_factory=dict
    )
    check_constraints: dict[str, CheckConstraintSnapshot] = Field(default_factory=dict)
    rls_enabled: bool = False
    rls_forced: bool = False

    class ConfigDict:
        populate_by_name = True


class EnumSnapshot(BaseModel):
    """Serialized PostgreSQL enum type."""

    name: str
    schema_name: str = Field(default="public", alias="schema")
    values: list[str]

    class ConfigDict:
        populate_by_name = True


class SequenceSnapshot(BaseModel):
    """Serialized sequence definition."""

    name: str
    schema_name: str = Field(default="public", alias="schema")
    start: int = 1
    increment: int = 1
    min_value: int | None = None
    max_value: int | None = None
    cache: int = 1
    cycle: bool = False
    owned_by: str | None = None  # "table.column" or None

    class ConfigDict:
        populate_by_name = True


class PolicyCommand(StrEnum):
    """RLS policy commands."""

    ALL = "all"
    SELECT = "select"
    INSERT = "insert"
    UPDATE = "update"
    DELETE = "delete"


class PolicySnapshot(BaseModel):
    """Serialized Row-Level Security policy."""

    name: str
    schema_name: str = Field(default="public", alias="schema")
    table: str
    command: PolicyCommand = PolicyCommand.ALL
    permissive: bool = True  # PERMISSIVE vs RESTRICTIVE
    roles: list[str] = Field(default_factory=lambda: ["public"])
    using: str | None = None  # USING expression
    with_check: str | None = None  # WITH CHECK expression

    class ConfigDict:
        populate_by_name = True
        use_enum_values = True


class RoleSnapshot(BaseModel):
    """Serialized database role."""

    name: str
    superuser: bool = False
    create_db: bool = False
    create_role: bool = False
    inherit: bool = True
    login: bool = False
    replication: bool = False
    bypass_rls: bool = False
    connection_limit: int = -1
    password: str | None = None  # Not stored in snapshots, just for creation
    valid_until: str | None = None
    in_roles: list[str] = Field(default_factory=list)


class GrantSnapshot(BaseModel):
    """Serialized grant/permission."""

    grantee: str  # Role name
    object_type: str  # "table", "schema", "sequence", "function"
    object_schema: str = "public"
    object_name: str
    privileges: list[str]  # ["SELECT", "INSERT", "UPDATE", "DELETE", "ALL"]
    with_grant_option: bool = False


class SchemaSnapshot(BaseModel):
    """Complete database schema snapshot.

    This is the root model that captures the entire database state.
    """

    version: SnapshotVersion = SnapshotVersion.V1
    dialect: Literal["postgresql"] = "postgresql"

    # Database objects
    tables: dict[str, TableSnapshot] = Field(default_factory=dict)
    enums: dict[str, EnumSnapshot] = Field(default_factory=dict)
    sequences: dict[str, SequenceSnapshot] = Field(default_factory=dict)
    schemas: Sequence[str] = Field(default=("public",))
    policies: dict[str, PolicySnapshot] = Field(default_factory=dict)
    roles: dict[str, RoleSnapshot] = Field(default_factory=dict)
    grants: Sequence[GrantSnapshot] = Field(default=())

    # Metadata for linking snapshots
    id: str = ""  # Unique snapshot identifier (e.g., "0001")
    prev_id: str | None = None  # Previous snapshot ID for diffing

    # Internal metadata for rename tracking (private attribute)
    _meta: dict = PrivateAttr(
        default_factory=lambda: {"schemas": {}, "tables": {}, "columns": {}}
    )

    class ConfigDict:
        use_enum_values = True

    def get_table_key(self, schema: str, table: str) -> str:
        """Get the dict key for a table."""
        if schema == "public":
            return table
        return f"{schema}.{table}"

    def get_enum_key(self, schema: str, name: str) -> str:
        """Get the dict key for an enum."""
        if schema == "public":
            return name
        return f"{schema}.{name}"

    def get_sequence_key(self, schema: str, name: str) -> str:
        """Get the dict key for a sequence."""
        if schema == "public":
            return name
        return f"{schema}.{name}"

    def get_policy_key(self, schema: str, table: str, name: str) -> str:
        """Get the dict key for a policy."""
        table_key = self.get_table_key(schema, table)
        return f"{table_key}.{name}"



"""Clerk authentication client."""

from __future__ import annotations

import uuid
from datetime import UTC, datetime
from typing import Any

from etils import epy

from derp.auth.base import BaseAuthClient
from derp.auth.models import AuthRequest, OrgInfo, OrgMemberInfo, SessionInfo, UserInfo
from derp.config import ClerkConfig

with epy.lazy_imports():
    import clerk_backend_api
    import clerk_backend_api.models.user as clerk_user_models
    from clerk_backend_api import security as clerk_security
    from clerk_backend_api.security import types as clerk_security_types


class ClerkAuthClient(BaseAuthClient):
    """Clerk-backed authentication client.

    Delegates authentication to Clerk's Backend API. Sign-up and sign-in
    are handled by Clerk's frontend SDK / hosted UI; this client provides
    server-side session validation, user lookup, and authorization.

    JWT verification is networkless — uses the PEM public key from
    ``ClerkConfig.jwt_key`` (Clerk Dashboard > API Keys > Show JWT public key).
    """

    def __init__(self, config: ClerkConfig):
        self._config = config
        self._clerk = clerk_backend_api.Clerk(bearer_auth=config.secret_key)

    def _to_user_info(self, user: Any) -> UserInfo:
        """Convert a Clerk SDK user object to UserInfo."""
        assert isinstance(user, clerk_user_models.User)
        email = ""
        if user.email_addresses:
            primary = next(
                (
                    e
                    for e in user.email_addresses
                    if e.id == user.primary_email_address_id
                ),
                user.email_addresses[0],
            )
            email = primary.email_address

        public_metadata = user.public_metadata if user.public_metadata else {}

        return UserInfo(
            id=user.id,
            email=email,
            first_name=user.first_name,
            last_name=user.last_name,
            username=user.username if user.username else None,
            image_url=user.image_url,
            role=public_metadata.get("role", "default"),
            is_active=not user.banned and not user.locked,
            is_superuser=public_metadata.get("is_superuser", False),
            created_at=datetime.fromtimestamp(user.created_at),
            updated_at=datetime.fromtimestamp(user.updated_at),
            last_sign_in_at=(
                datetime.fromtimestamp(user.last_sign_in_at)
                if user.last_sign_in_at
                else None
            ),
            email_confirmed_at=None,
            metadata={
                "public_metadata": public_metadata,
                "private_metadata": (
                    user.private_metadata if user.private_metadata else {}
                ),
                "banned": user.banned,
                "locked": user.locked,
            },
        )

    # ------------------------------------------------------------------
    # User management
    # ------------------------------------------------------------------

    async def get_user(self, user_id: str | uuid.UUID) -> UserInfo | None:
        """Get a user by their Clerk user ID."""
        user = await self._clerk.users.get_async(user_id=str(user_id))
        return self._to_user_info(user)

    async def list_users(
        self, *, limit: int | None = None, offset: int | None = None
    ) -> list[UserInfo]:
        """List Clerk users ordered by creation date (newest first)."""
        kwargs: dict[str, Any] = {"order_by": "-created_at"}
        if limit is not None:
            kwargs["limit"] = limit
        if offset is not None:
            kwargs["offset"] = offset
        users = await self._clerk.users.list_async(**kwargs)
        return [self._to_user_info(u) for u in users]

    async def update_user(
        self,
        *,
        user_id: str | uuid.UUID,
        email: str | None = None,
        **kwargs: Any,
    ) -> UserInfo:
        """Update a Clerk user.

        Supports ``first_name``, ``last_name``, ``username`` as direct kwargs.
        All other kwargs are merged into ``public_metadata``.
        """
        if email is not None:
            raise NotImplementedError(
                "Email updates are not supported via update_user for Clerk. "
                "Use Clerk's email_addresses API instead."
            )

        params: dict[str, Any] = {}

        for field_name in ("first_name", "last_name", "username"):
            if field_name in kwargs:
                params[field_name] = kwargs.pop(field_name)

        if kwargs:
            params["public_metadata"] = kwargs

        user = await self._clerk.users.update_async(
            user_id=str(user_id),
            **params,
        )
        return self._to_user_info(user)

    async def delete_user(self, user_id: str | uuid.UUID) -> None:
        """Delete a Clerk user."""
        await self._clerk.users.delete_async(user_id=str(user_id))

    async def count_users(self) -> int:
        """Return the total number of Clerk users."""
        return await self._clerk.users.count_async()

    # ------------------------------------------------------------------
    # Session validation
    # ------------------------------------------------------------------

    async def authenticate(self, request: AuthRequest) -> SessionInfo | None:
        """Authenticate a request via Clerk (networkless).

        Uses the Clerk SDK's ``authenticate_request`` with the PEM
        public key from ``ClerkConfig.jwt_key``. Returns ``SessionInfo``
        if valid, ``None`` otherwise.
        """
        options = clerk_security.AuthenticateRequestOptions(
            jwt_key=self._config.jwt_key,
            authorized_parties=(
                list(self._config.authorized_parties)
                if self._config.authorized_parties
                else None
            ),
        )
        result = clerk_security.authenticate_request(request, options)
        if not result.is_signed_in or result.payload is None:
            return None

        auth = result.to_auth()
        if isinstance(auth, clerk_security_types.SessionAuthObjectV2):
            return SessionInfo(
                user_id=auth.sub or "",
                session_id=auth.sid or "",
                role=auth.org_role or auth.role or "default",
                expires_at=(
                    datetime.fromtimestamp(auth.exp, tz=UTC)
                    if auth.exp is not None
                    else datetime.now(UTC)
                ),
                metadata=result.payload or {},
                org_id=getattr(auth, "org_id", None),
                org_role=auth.org_role or None,
            )
        elif isinstance(auth, clerk_security_types.SessionAuthObjectV1):
            claims = auth.claims if auth.claims is not None else {}
            return SessionInfo(
                user_id=auth.user_id or "",
                session_id=auth.session_id or "",
                role=auth.org_role or claims.get("role", "default"),
                expires_at=(
                    datetime.fromtimestamp(claims["exp"], tz=UTC)
                    if "exp" in claims
                    else datetime.now(UTC)
                ),
                metadata=result.payload or {},
                org_id=getattr(auth, "org_id", None),
                org_role=auth.org_role or None,
            )
        raise ValueError(f"Unknown auth object type: {type(auth)}")

    async def list_sessions(
        self,
        *,
        user_id: str | uuid.UUID | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[Any]:
        """List active Clerk sessions."""
        kwargs: dict[str, Any] = {"status": "active"}
        if user_id is not None:
            kwargs["user_id"] = str(user_id)
        if limit is not None:
            kwargs["limit"] = limit
        if offset is not None:
            kwargs["offset"] = offset
        return await self._clerk.sessions.list_async(**kwargs)

    # ------------------------------------------------------------------
    # Sign-out
    # ------------------------------------------------------------------

    async def sign_out(self, session_id: str | uuid.UUID) -> None:
        """Revoke a Clerk session."""
        await self._clerk.sessions.revoke_async(session_id=str(session_id))

    async def sign_out_all(self, user_id: str | uuid.UUID) -> None:
        """Revoke all active sessions for a user."""
        sessions = await self._clerk.sessions.list_async(
            user_id=str(user_id), status="active"
        )
        if sessions:
            for session in sessions:
                await self._clerk.sessions.revoke_async(session_id=session.id)

    # ------------------------------------------------------------------
    # Organizations
    # ------------------------------------------------------------------

    def _to_org_info(self, org: Any) -> OrgInfo:
        """Convert a Clerk SDK organization object to OrgInfo."""
        return OrgInfo(
            id=org.id,
            name=org.name,
            slug=org.slug or "",
            metadata=org.public_metadata if org.public_metadata else {},
            created_at=datetime.fromtimestamp(org.created_at),
            updated_at=datetime.fromtimestamp(org.updated_at),
        )

    def _to_org_member_info(self, member: Any) -> OrgMemberInfo:
        """Convert a Clerk SDK organization membership to OrgMemberInfo."""
        return OrgMemberInfo(
            org_id=member.organization.id if member.organization else "",
            user_id=member.public_user_data.user_id if member.public_user_data else "",
            role=member.role or "member",
            created_at=datetime.fromtimestamp(member.created_at),
            updated_at=datetime.fromtimestamp(member.updated_at),
        )

    async def create_org(
        self,
        *,
        name: str,
        slug: str,
        creator_id: str | uuid.UUID,
        **kwargs: Any,
    ) -> OrgInfo:
        """Create a Clerk organization."""
        org = await self._clerk.organizations.create_async(
            request={
                "name": name,
                "slug": slug,
                "created_by": str(creator_id),
            },
        )
        return self._to_org_info(org)

    async def get_org(self, org_id: str | uuid.UUID) -> OrgInfo | None:
        """Get a Clerk organization by ID."""
        org = await self._clerk.organizations.get_async(
            organization_id=str(org_id),
        )
        return self._to_org_info(org) if org else None

    async def get_org_by_slug(self, slug: str) -> OrgInfo | None:
        """Get a Clerk organization by slug."""
        orgs = await self._clerk.organizations.list_async(query=slug, limit=1)
        for org in orgs:
            if org.slug == slug:
                return self._to_org_info(org)
        return None

    async def update_org(
        self,
        *,
        org_id: str | uuid.UUID,
        name: str | None = None,
        slug: str | None = None,
        **kwargs: Any,
    ) -> OrgInfo:
        """Update a Clerk organization."""
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if slug is not None:
            params["slug"] = slug
        org = await self._clerk.organizations.update_async(
            organization_id=str(org_id),
            **params,
        )
        return self._to_org_info(org)

    async def delete_org(self, org_id: str | uuid.UUID) -> None:
        """Delete a Clerk organization."""
        await self._clerk.organizations.delete_async(
            organization_id=str(org_id),
        )

    async def list_orgs(
        self,
        *,
        user_id: str | uuid.UUID | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[OrgInfo]:
        """List Clerk organizations."""
        kwargs: dict[str, Any] = {}
        if user_id is not None:
            kwargs["user_id"] = [str(user_id)]
        if limit is not None:
            kwargs["limit"] = limit
        if offset is not None:
            kwargs["offset"] = offset
        orgs = await self._clerk.organizations.list_async(**kwargs)
        return [self._to_org_info(o) for o in orgs]

    # ------------------------------------------------------------------
    # Organization membership
    # ------------------------------------------------------------------

    async def add_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
        role: str = "member",
    ) -> OrgMemberInfo:
        """Add a user to a Clerk organization."""
        member = await self._clerk.organization_memberships.create_async(
            organization_id=str(org_id),
            user_id=str(user_id),
            role=role,
        )
        return self._to_org_member_info(member)

    async def update_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
        role: str,
    ) -> OrgMemberInfo:
        """Update a member's role in a Clerk organization."""
        member = await self._clerk.organization_memberships.update_async(
            organization_id=str(org_id),
            user_id=str(user_id),
            role=role,
        )
        return self._to_org_member_info(member)

    async def remove_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
    ) -> None:
        """Remove a user from a Clerk organization."""
        await self._clerk.organization_memberships.delete_async(
            organization_id=str(org_id),
            user_id=str(user_id),
        )

    async def list_org_members(
        self,
        org_id: str | uuid.UUID,
        *,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[OrgMemberInfo]:
        """List members of a Clerk organization."""
        kwargs: dict[str, Any] = {}
        if limit is not None:
            kwargs["limit"] = limit
        if offset is not None:
            kwargs["offset"] = offset
        members = await self._clerk.organization_memberships.list_async(
            organization_id=str(org_id),
            **kwargs,
        )
        return [self._to_org_member_info(m) for m in members]

    async def get_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
    ) -> OrgMemberInfo | None:
        """Get a single membership from a Clerk organization."""
        members = await self._clerk.organization_memberships.list_async(
            organization_id=str(org_id),
        )
        for m in members:
            uid = m.public_user_data.user_id if m.public_user_data else None
            if uid == str(user_id):
                return self._to_org_member_info(m)
        return None



"""S3-compatible storage client using aiobotocore."""

from __future__ import annotations

from types import TracebackType
from typing import Any

import aiobotocore.session
from aiobotocore.client import AioBaseClient
from botocore.config import Config
from botocore.exceptions import ClientError

from derp.config import StorageConfig


class StorageClient:
    """S3-compatible storage client for uploading and fetching files.

    Example::

        config = StorageConfig(
            endpoint_url="https://s3.amazonaws.com",
            access_key_id="key",
            secret_access_key="secret",
        )
        storage = StorageClient(config)
        await storage.connect()
        await storage.upload_file("local.txt", "remote.txt")
        await storage.disconnect()
    """

    def __init__(self, config: StorageConfig):
        """Initialize Storage client.

        Args:
            config: Storage configuration.
        """
        self._config = config
        self._session: aiobotocore.session.AioSession | None = None
        self._client: AioBaseClient | None = None

    async def connect(self) -> None:
        """Establish connection to S3."""
        if self._client is not None:
            return

        self._session = aiobotocore.session.get_session()

        config = Config(
            region_name=self._config.region,
            signature_version="s3v4",
        )

        self._client = await self._session.create_client(
            self._config.service_name,
            endpoint_url=self._config.endpoint_url,
            aws_access_key_id=self._config.access_key_id,
            aws_secret_access_key=self._config.secret_access_key,
            aws_session_token=self._config.session_token,
            use_ssl=self._config.use_ssl,
            verify=self._config.verify,
            config=config,
        ).__aenter__()

    async def disconnect(self) -> None:
        """Close connection to S3."""
        if self._client is not None:
            await self._client.__aexit__(None, None, None)
            self._client = None
        if self._session is not None:
            self._session = None

    async def __aenter__(self) -> StorageClient:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.disconnect()

    @property
    def client(self) -> AioBaseClient:
        """Get the S3 client."""
        if self._client is None:
            raise RuntimeError(
                "Storage not connected. Call connect() or use async context manager."
            )
        return self._client

    def get_url(self, *, bucket: str, key: str) -> str:
        """Get the URL for a file in S3.

        Args:
            bucket: Name of the S3 bucket.
            key: S3 object key (path in bucket).

        Returns:
            URL for the file.

        Raises:
            ValueError: If endpoint_url is not configured.
        """
        if self._config.endpoint_url is None:
            raise ValueError(
                "Cannot construct URL: `endpoint_url` is not "
                "configured in StorageConfig."
            )
        return f"{self._config.endpoint_url}/{bucket}/{key}"

    async def upload_file(
        self,
        *,
        bucket: str,
        key: str,
        data: bytes,
        content_type: str | None = None,
        metadata: dict[str, str] | None = None,
        extra_args: dict[str, Any] | None = None,
    ) -> None:
        """Upload a file to S3.

        Args:
            bucket: Name of the S3 bucket.
            key: S3 object key (path in bucket).
            data: Bytes to upload.
            content_type: MIME type of the file.
            metadata: Metadata to attach to the object.
            extra_args: Additional arguments to pass to put_object.

        Example::

            await storage.upload_file(
                bucket="my-bucket",
                key="remote/file.txt",
                data=b"Hello, World!",
                content_type="text/plain",
                metadata={"author": "user123"},
            )
        """
        put_kwargs: dict[str, Any] = {
            "Bucket": bucket,
            "Key": key,
            "Body": data,
        }

        if content_type:
            put_kwargs["ContentType"] = content_type

        if metadata:
            put_kwargs["Metadata"] = metadata

        if extra_args:
            put_kwargs.update(extra_args)

        await self.client.put_object(**put_kwargs)

    async def fetch_file(self, *, bucket: str, key: str) -> bytes:
        """Fetch a file from S3.

        Args:
            bucket: Name of the S3 bucket.
            key: S3 object key (path in bucket).

        Returns:
            File content as bytes

        Example:
            # Get content as bytes
            content = await storage.fetch_file("remote/file.txt")

            # Save to local file
            await storage.fetch_file("remote/file.txt")
        """
        response = await self.client.get_object(
            Bucket=bucket,
            Key=key,
        )

        async with response["Body"] as stream:
            content = await stream.read()

        return content

    async def delete_file(self, *, bucket: str, key: str) -> None:
        """Delete a file from S3.

        Args:
            key: S3 object key (path in bucket)

        Example:
            await storage.delete_file("remote/file.txt")
        """
        await self.client.delete_object(Bucket=bucket, Key=key)

    async def file_exists(self, *, bucket: str, key: str) -> bool:
        """Check if a file exists in S3.

        Args:
            key: S3 object key (path in bucket)

        Returns:
            True if file exists, False otherwise

        Example:
            exists = await storage.file_exists("remote/file.txt")
        """
        try:
            await self.client.head_object(Bucket=bucket, Key=key)
            return True
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code", "")
            if error_code in ("404", "NoSuchKey"):
                return False
            # Re-raise other client errors
            raise

    async def head_object(self, *, bucket: str, key: str) -> dict[str, Any]:
        """Get object metadata without downloading the content.

        Args:
            bucket: Name of the S3 bucket.
            key: S3 object key (path in bucket).

        Returns:
            Dict with 'content_type', 'content_length', 'last_modified',
            'etag', and 'metadata' keys.
        """
        response = await self.client.head_object(Bucket=bucket, Key=key)
        return {
            "content_type": response.get("ContentType", "application/octet-stream"),
            "content_length": response.get("ContentLength", 0),
            "last_modified": response["LastModified"].isoformat(),
            "etag": response.get("ETag", "").strip('"'),
            "metadata": response.get("Metadata", {}),
        }

    async def list_files(
        self, *, bucket: str, prefix: str = "", max_keys: int | None = None
    ) -> list[str]:
        """List files in S3 bucket.

        Args:
            bucket: Name of the S3 bucket.
            prefix: Prefix to filter files by.
            max_keys: Maximum number of keys to return.

        Returns:
            List of object keys.

        Example:
            files = await storage.list_files(bucket="my-bucket", prefix="folder/")
        """
        list_kwargs: dict[str, Any] = {
            "Bucket": bucket,
        }

        if prefix:
            list_kwargs["Prefix"] = prefix

        if max_keys:
            list_kwargs["MaxKeys"] = max_keys

        response = await self.client.list_objects_v2(**list_kwargs)

        if "Contents" not in response:
            return []

        return [obj["Key"] for obj in response["Contents"]]

    async def list_buckets(self) -> list[dict[str, Any]]:
        """List all S3 buckets.

        Returns:
            List of dicts with 'name' and 'creation_date' keys.
        """
        response = await self.client.list_buckets()
        return [
            {
                "name": b["Name"],
                "creation_date": b["CreationDate"].isoformat(),
            }
            for b in response.get("Buckets", [])
        ]

    async def list_objects(
        self,
        *,
        bucket: str,
        prefix: str = "",
        max_keys: int = 1000,
    ) -> dict[str, Any]:
        """List objects and common prefixes in a bucket.

        Uses ``Delimiter='/'`` for folder-like browsing.

        Args:
            bucket: Name of the S3 bucket.
            prefix: Prefix to filter by (use trailing ``/`` for folders).
            max_keys: Maximum number of keys to return.

        Returns:
            Dict with 'objects' (list of object metadata dicts) and
            'prefixes' (list of prefix strings representing folders).
        """
        list_kwargs: dict[str, Any] = {
            "Bucket": bucket,
            "Prefix": prefix,
            "Delimiter": "/",
            "MaxKeys": max_keys,
        }

        response = await self.client.list_objects_v2(**list_kwargs)

        objects = [
            {
                "key": obj["Key"],
                "size": obj["Size"],
                "last_modified": obj["LastModified"].isoformat(),
            }
            for obj in response.get("Contents", [])
            if obj["Key"] != prefix
        ]

        prefixes = [p["Prefix"] for p in response.get("CommonPrefixes", [])]

        return {"objects": objects, "prefixes": prefixes}



"""Studio command - launch the Derp Studio web interface."""

from __future__ import annotations

import os
import shutil
import subprocess
import time
from pathlib import Path
from typing import Annotated

import typer
import uvicorn

from derp.config import ConfigError, DerpConfig
from derp.studio.server import create_app


def _validate_config() -> None:
    try:
        DerpConfig.load()
    except ConfigError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(1)


def _studio_ui_dir() -> Path:
    return Path(__file__).resolve().parents[2] / "studio" / "ui"


def _terminate_process(process: subprocess.Popen[bytes]) -> None:
    if process.poll() is not None:
        return

    process.terminate()
    try:
        process.wait(timeout=5)
    except subprocess.TimeoutExpired:
        process.kill()
        process.wait(timeout=5)


def studio(
    host: Annotated[
        str, typer.Option("--host", "-h", help="Host to bind to")
    ] = "127.0.0.1",
    port: Annotated[int, typer.Option("--port", "-p", help="Port to bind to")] = 4983,
) -> None:
    """Launch Derp Studio - a web UI for browsing your database."""
    _validate_config()

    old_url = os.environ.pop("PUBLIC_API_URL", None)
    old_node_env = os.environ.pop("NODE_ENV", None)
    os.environ["PUBLIC_API_URL"] = f"http://{host}:{port}"
    os.environ["NODE_ENV"] = "production"

    studio_app = create_app()
    uvicorn.run(studio_app, host=host, port=port)
    if old_url is not None:
        os.environ["PUBLIC_API_URL"] = old_url
    if old_node_env is not None:
        os.environ["NODE_ENV"] = old_node_env


def studio_dev(
    host: Annotated[
        str, typer.Option("--host", help="Host to bind both servers")
    ] = "127.0.0.1",
    backend_port: Annotated[
        int, typer.Option("--backend-port", help="Backend API port")
    ] = 4983,
    frontend_port: Annotated[
        int, typer.Option("--frontend-port", help="Frontend Vite dev port")
    ] = 5173,
) -> None:
    """Launch Derp Studio in development mode (backend + frontend)."""
    _validate_config()

    bun = shutil.which("bun")
    if bun is None:
        typer.echo(
            "Error: `bun` is required for `derp studio-dev` but was not found on PATH.",
            err=True,
        )
        typer.echo("Run `./scripts/build_studio.sh` after installing Bun.", err=True)
        raise typer.Exit(1)

    ui_dir = _studio_ui_dir()
    if not ui_dir.exists():
        typer.echo(f"Error: Studio UI directory not found: {ui_dir}", err=True)
        raise typer.Exit(1)

    frontend_cmd = [
        bun,
        "run",
        "dev",
        "--",
        "--host",
        host,
        "--port",
        str(frontend_port),
        "--strictPort",
    ]
    api_origin = f"http://{host}:{backend_port}"
    frontend_origin = f"http://{host}:{frontend_port}"
    env = os.environ.copy()
    env["NODE_ENV"] = "development"
    env["PUBLIC_API_URL"] = api_origin

    frontend_process: subprocess.Popen[bytes] | None = None
    try:
        frontend_process = subprocess.Popen(
            frontend_cmd,
            cwd=ui_dir,
            env=env,
        )

        time.sleep(0.3)
        return_code = frontend_process.poll()
        if return_code is not None:
            typer.echo(
                (
                    "Error: Frontend dev server exited early with code "
                    f"{return_code}. Run `./scripts/build_studio.sh`."
                ),
                err=True,
            )
            raise typer.Exit(1)

        typer.echo(f"Studio frontend: {frontend_origin}")
        typer.echo(f"Studio backend:  {api_origin}/api/config")
        typer.echo(f"Studio app URL:  {api_origin}")

        studio_app = create_app()
        uvicorn.run(studio_app, host=host, port=backend_port)
    finally:
        if frontend_process is not None:
            _terminate_process(frontend_process)



"""Celery queue client."""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from datetime import timedelta
from typing import Any

from etils import epy

from derp.config import CeleryConfig
from derp.queue.base import QueueClient, Schedule, ScheduleType, TaskState, TaskStatus
from derp.queue.exceptions import QueueNotConnectedError, QueueProviderError

with epy.lazy_imports():
    import celery
    import celery.result as celery_result
    import celery.schedules as celery_schedules


# Map Celery states to TaskState.
_CELERY_STATE_MAP: dict[str, TaskState] = {
    "PENDING": TaskState.PENDING,
    "STARTED": TaskState.STARTED,
    "SUCCESS": TaskState.SUCCESS,
    "FAILURE": TaskState.FAILURE,
    "REVOKED": TaskState.REVOKED,
    "RETRY": TaskState.STARTED,
    "RECEIVED": TaskState.PENDING,
}


class CeleryQueueClient(QueueClient):
    """Queue client backed by Celery."""

    supports_result = True
    supports_revoke = True
    supports_delay = True

    def __init__(self, config: CeleryConfig):
        self._config = config
        self._app: celery.Celery | None = None
        self._schedules: list[Schedule] = []

    @property
    def app(self) -> celery.Celery:
        """Expose the underlying Celery app for worker-side task registration."""
        if self._app is None:
            self._create_app()
        return self._app  # type: ignore[return-value]

    def _create_app(self) -> None:
        self._app = celery.Celery("derp")
        self._app.conf.update(
            broker_url=self._config.broker_url,
            result_backend=self._config.result_backend,
            task_serializer=self._config.task_serializer,
            result_serializer=self._config.result_serializer,
            task_default_queue=self._config.task_default_queue,
        )

    async def connect(self) -> None:
        if self._app is not None:
            return
        self._create_app()
        if self._schedules:
            self._apply_beat_schedule()

    async def disconnect(self) -> None:
        if self._app is not None:
            self._app.close()
            self._app = None

    async def enqueue(
        self,
        task_name: str,
        payload: dict[str, Any] | None = None,
        *,
        queue: str | None = None,
        delay: timedelta | None = None,
    ) -> str:
        if self._app is None:
            raise QueueNotConnectedError()

        kwargs: dict[str, Any] = {}
        if queue is not None:
            kwargs["queue"] = queue
        if delay is not None:
            kwargs["countdown"] = delay.total_seconds()

        try:
            result = await asyncio.to_thread(
                self._app.send_task,
                task_name,
                kwargs=payload,
                **kwargs,
            )
        except Exception as exc:
            raise QueueProviderError(str(exc) or "Failed to enqueue task") from exc

        return str(result.id)

    async def get_status(self, task_id: str) -> TaskStatus:
        if self._app is None:
            raise QueueNotConnectedError()

        def _fetch_status() -> tuple[str, Any]:
            r = celery_result.AsyncResult(task_id, app=self._app)
            return r.state, r.result

        try:
            raw_state, raw_result = await asyncio.to_thread(_fetch_status)
            state = _CELERY_STATE_MAP.get(raw_state, TaskState.UNKNOWN)
            task_result = raw_result if state == TaskState.SUCCESS else None
            error = str(raw_result) if state == TaskState.FAILURE else None
        except Exception as exc:
            raise QueueProviderError(str(exc) or "Failed to get task status") from exc

        return TaskStatus(
            task_id=task_id,
            state=state,
            result=task_result,
            error=error,
        )

    def register_schedules(self, schedules: Sequence[Schedule]) -> None:
        """Register recurring schedules with Celery Beat."""
        self._schedules = list(schedules)
        if self._app is not None:
            self._apply_beat_schedule()

    def get_schedules(self) -> list[Schedule]:
        """Return the currently registered schedules."""
        return self._schedules

    def _apply_beat_schedule(self) -> None:
        """Write schedules into Celery's beat_schedule config."""
        assert self._app is not None
        beat_schedule: dict[str, Any] = {}
        for s in self._schedules:
            entry: dict[str, Any] = {"task": s.task}
            if s.type == ScheduleType.CRON and s.cron:
                parts = s.cron.split()
                entry["schedule"] = celery_schedules.crontab(
                    minute=parts[0] if len(parts) > 0 else "*",
                    hour=parts[1] if len(parts) > 1 else "*",
                    day_of_month=parts[2] if len(parts) > 2 else "*",
                    month_of_year=parts[3] if len(parts) > 3 else "*",
                    day_of_week=parts[4] if len(parts) > 4 else "*",
                )
            elif s.type == ScheduleType.INTERVAL and s.interval:
                entry["schedule"] = s.interval.total_seconds()
            if s.payload:
                entry["kwargs"] = s.payload
            if s.queue:
                entry["options"] = {"queue": s.queue}
            beat_schedule[s.name] = entry
        self._app.conf.beat_schedule = beat_schedule



"""Role and grant operation convertors."""

from __future__ import annotations

from derp.orm.migrations.convertors.base import (
    ConvertorRegistry,
    StatementConvertor,
    quote_identifier,
    quote_schema_table,
)
from derp.orm.migrations.statements.types import (
    CreateRoleStatement,
    DropRoleStatement,
    GrantStatement,
    RevokeStatement,
)


class CreateRoleConvertor(StatementConvertor[CreateRoleStatement]):
    """Convert CREATE ROLE statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "create_role"

    def convert(self, statement: CreateRoleStatement) -> str:
        role_name = quote_identifier(statement.name)
        options: list[str] = []

        if statement.superuser:
            options.append("SUPERUSER")
        else:
            options.append("NOSUPERUSER")

        if statement.create_db:
            options.append("CREATEDB")
        else:
            options.append("NOCREATEDB")

        if statement.create_role:
            options.append("CREATEROLE")
        else:
            options.append("NOCREATEROLE")

        if statement.inherit:
            options.append("INHERIT")
        else:
            options.append("NOINHERIT")

        if statement.login:
            options.append("LOGIN")
        else:
            options.append("NOLOGIN")

        if statement.replication:
            options.append("REPLICATION")
        else:
            options.append("NOREPLICATION")

        if statement.bypass_rls:
            options.append("BYPASSRLS")
        else:
            options.append("NOBYPASSRLS")

        if statement.connection_limit != -1:
            options.append(f"CONNECTION LIMIT {statement.connection_limit}")

        if statement.password:
            # Note: In production, use a safer method for passwords
            options.append(f"PASSWORD '{statement.password}'")

        if statement.valid_until:
            options.append(f"VALID UNTIL '{statement.valid_until}'")

        sql = f"CREATE ROLE {role_name}"
        if options:
            sql += " WITH " + " ".join(options)
        sql += ";"

        # IN ROLE (member of other roles)
        if statement.in_roles:
            for parent_role in statement.in_roles:
                sql += f"\nGRANT {quote_identifier(parent_role)} TO {role_name};"

        return sql


class DropRoleConvertor(StatementConvertor[DropRoleStatement]):
    """Convert DROP ROLE statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "drop_role"

    def convert(self, statement: DropRoleStatement) -> str:
        role_name = quote_identifier(statement.name)
        return f"DROP ROLE IF EXISTS {role_name};"


class GrantConvertor(StatementConvertor[GrantStatement]):
    """Convert GRANT statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "grant"

    def convert(self, statement: GrantStatement) -> str:
        privileges = ", ".join(statement.privileges)
        grantee = quote_identifier(statement.grantee)

        # Build object reference
        if statement.object_type.upper() == "TABLE":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON TABLE {obj_ref}"
        elif statement.object_type.upper() == "SCHEMA":
            on_clause = f"ON SCHEMA {quote_identifier(statement.object_name)}"
        elif statement.object_type.upper() == "SEQUENCE":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON SEQUENCE {obj_ref}"
        elif statement.object_type.upper() == "FUNCTION":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON FUNCTION {obj_ref}"
        else:
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON {statement.object_type.upper()} {obj_ref}"

        sql = f"GRANT {privileges} {on_clause} TO {grantee}"

        if statement.with_grant_option:
            sql += " WITH GRANT OPTION"

        return sql + ";"


class RevokeConvertor(StatementConvertor[RevokeStatement]):
    """Convert REVOKE statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "revoke"

    def convert(self, statement: RevokeStatement) -> str:
        privileges = ", ".join(statement.privileges)
        grantee = quote_identifier(statement.grantee)

        # Build object reference
        if statement.object_type.upper() == "TABLE":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON TABLE {obj_ref}"
        elif statement.object_type.upper() == "SCHEMA":
            on_clause = f"ON SCHEMA {quote_identifier(statement.object_name)}"
        elif statement.object_type.upper() == "SEQUENCE":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON SEQUENCE {obj_ref}"
        elif statement.object_type.upper() == "FUNCTION":
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON FUNCTION {obj_ref}"
        else:
            obj_ref = quote_schema_table(statement.object_schema, statement.object_name)
            on_clause = f"ON {statement.object_type.upper()} {obj_ref}"

        sql = f"REVOKE {privileges} {on_clause} FROM {grantee}"

        if statement.cascade:
            sql += " CASCADE"

        return sql + ";"


# Register convertors
ConvertorRegistry.register(CreateRoleConvertor())
ConvertorRegistry.register(DropRoleConvertor())
ConvertorRegistry.register(GrantConvertor())
ConvertorRegistry.register(RevokeConvertor())



"""Drizzle-style migration infrastructure for Derp ORM."""

from derp.orm.migrations.journal import JournalEntry, MigrationJournal
from derp.orm.migrations.snapshot.models import (
    CheckConstraintSnapshot,
    ColumnSnapshot,
    EnumSnapshot,
    ForeignKeySnapshot,
    IdentityConfig,
    IndexSnapshot,
    PolicySnapshot,
    RoleSnapshot,
    SchemaSnapshot,
    SequenceSnapshot,
    TableSnapshot,
    UniqueConstraintSnapshot,
)

__all__ = [
    # Journal
    "JournalEntry",
    "MigrationJournal",
    # Snapshot models
    "SchemaSnapshot",
    "TableSnapshot",
    "ColumnSnapshot",
    "ForeignKeySnapshot",
    "IndexSnapshot",
    "UniqueConstraintSnapshot",
    "CheckConstraintSnapshot",
    "EnumSnapshot",
    "SequenceSnapshot",
    "PolicySnapshot",
    "RoleSnapshot",
    "IdentityConfig",
]



"""Generate migration command."""

from __future__ import annotations

from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from typing import Annotated

import typer

from derp.config import ConfigError, DerpConfig
from derp.orm.loader import discover_tables

# Import all convertors to register them
from derp.orm.migrations.convertors import (  # noqa: F401
    column,
    constraint,
    enum,
    index,
    policy,
    role,
    schema,
    sequence,
    table,
)
from derp.orm.migrations.convertors.base import ConvertorRegistry
from derp.orm.migrations.journal import (
    get_next_version,
    load_journal,
    load_latest_snapshot,
    save_journal,
    save_migration_sql,
    save_snapshot,
)
from derp.orm.migrations.safety import (
    detect_destructive_operations,
    format_destructive_warnings,
    has_high_risk_operations,
)
from derp.orm.migrations.snapshot.differ import SnapshotDiffer
from derp.orm.migrations.snapshot.models import ColumnSnapshot, SchemaSnapshot
from derp.orm.migrations.snapshot.serializer import serialize_schema


def _columns_match(old_col: ColumnSnapshot, new_col: ColumnSnapshot) -> bool:
    """Check if two columns are similar enough to be a rename candidate."""
    return (
        old_col.type == new_col.type
        and old_col.not_null == new_col.not_null
        and old_col.default == new_col.default
    )


def _find_rename_candidates(
    old_snapshot: SchemaSnapshot, new_snapshot: SchemaSnapshot
) -> list[tuple[str, str, str, str]]:
    """Find potential column renames between snapshots.

    Returns list of (table_name, old_col, new_col, col_type) tuples.
    """
    candidates: list[tuple[str, str, str, str]] = []

    # Find tables that exist in both snapshots
    common_tables = set(old_snapshot.tables.keys()) & set(new_snapshot.tables.keys())

    for table_key in common_tables:
        old_table = old_snapshot.tables[table_key]
        new_table = new_snapshot.tables[table_key]

        old_cols = set(old_table.columns.keys())
        new_cols = set(new_table.columns.keys())

        dropped = old_cols - new_cols
        added = new_cols - old_cols

        # Find matching pairs
        for old_name in sorted(dropped):
            old_col = old_table.columns[old_name]
            for new_name in sorted(added):
                new_col = new_table.columns[new_name]
                if _columns_match(old_col, new_col):
                    candidates.append(
                        (old_table.name, old_name, new_name, old_col.type)
                    )

    return candidates


def create_rename_resolver(
    old_snapshot: SchemaSnapshot,
    new_snapshot: SchemaSnapshot,
    force: bool = False,
) -> dict[str, str]:
    """Prompt user for potential renames and return decisions.

    Returns a dict mapping "table.old_col" -> "new_col" for confirmed renames.
    """
    if force:
        # In force mode, skip prompts and treat as drop+add (safe default)
        return {}

    candidates = _find_rename_candidates(old_snapshot, new_snapshot)
    if not candidates:
        return {}

    decisions: dict[str, str] = {}
    used_pairs: set[tuple[str, str, str]] = set()  # (table, old, new) already decided

    for table_name, old_name, new_name, col_type in candidates:
        # Skip if either column is already part of a confirmed rename
        key = f"{table_name}.{old_name}"
        if key in decisions:
            continue

        # Check if either column already matched in a different pair
        skip = False
        for t, o, n in used_pairs:
            if t == table_name and (o == old_name or n == new_name):
                skip = True
                break
        if skip:
            continue

        typer.echo("")
        typer.echo(f"Potential rename detected in table '{table_name}':")
        typer.echo(f"  Column '{old_name}' ({col_type}) was removed")
        typer.echo(f"  Column '{new_name}' ({col_type}) was added")

        prompt = f"Did you rename '{old_name}' to '{new_name}'?"
        if typer.confirm(prompt, default=False):
            decisions[key] = new_name
            used_pairs.add((table_name, old_name, new_name))

    return decisions


def make_rename_callback(
    decisions: dict[str, str],
) -> Callable[[str, str, str], bool]:
    """Create a callback function for the differ from user decisions."""

    def resolver(object_type: str, old_name: str, new_name: str) -> bool:
        if object_type != "column":
            return False
        # old_name is "table.column" format
        return decisions.get(old_name) == new_name

    return resolver


def generate(
    name: Annotated[
        str, typer.Option("--name", "-n", help="Migration name")
    ] = "migration",
    force: Annotated[
        bool,
        typer.Option("--force", "-f", help="Skip confirmation for destructive changes"),
    ] = False,
    custom: Annotated[
        bool, typer.Option("--custom", help="Generate empty migration for custom SQL")
    ] = False,
) -> None:
    """Generate a new migration from schema changes.

    Compares your current schema definition against the latest snapshot
    and generates migration SQL with a new snapshot file.
    """
    try:
        config = DerpConfig.load()
    except ConfigError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(1)

    migrations_dir = Path(config.database.migrations_dir)
    schema_path = config.database.schema_path

    if custom:
        _generate_custom_migration(config, name)
        return

    # Load tables from schema module
    try:
        tables = discover_tables(
            schema_path,
            include_auth=config.auth is not None and config.auth.native is not None,
        )
    except FileNotFoundError:
        typer.echo(f"Error: Schema file not found: {schema_path}", err=True)
        raise typer.Exit(1)

    if not tables:
        typer.echo(f"No Table classes found in {schema_path}", err=True)
        raise typer.Exit(1)

    typer.echo(f"Found {len(tables)} table(s): {', '.join(t.__name__ for t in tables)}")

    # Load journal and previous snapshot
    journal = load_journal(migrations_dir)
    prev_snapshot_data = load_latest_snapshot(migrations_dir, journal)

    if prev_snapshot_data:
        prev_snapshot = SchemaSnapshot.model_validate(prev_snapshot_data)
        typer.echo(f"Previous snapshot: {journal.get_latest_version()}")
    else:
        prev_snapshot = SchemaSnapshot()
        typer.echo("No previous snapshot found, creating initial migration")

    # Serialize current schema to snapshot
    next_version = get_next_version(journal)
    current_snapshot = serialize_schema(
        tables,
        schema="public",
        snapshot_id=next_version,
        prev_id=prev_snapshot.id if prev_snapshot.id else None,
    )

    # Prompt for potential column renames before diffing
    rename_decisions = create_rename_resolver(prev_snapshot, current_snapshot, force)
    rename_callback = (
        make_rename_callback(rename_decisions) if rename_decisions else None
    )

    # Diff snapshots
    differ = SnapshotDiffer(prev_snapshot, current_snapshot, rename_callback)
    statements = differ.diff()

    if not statements:
        typer.echo("No changes detected. Schema is up to date.")
        return

    typer.echo(f"Detected {len(statements)} change(s)")

    # Check for destructive operations
    destructive = detect_destructive_operations(statements)
    if destructive:
        typer.echo("")
        typer.echo(format_destructive_warnings(destructive), err=True)

        if has_high_risk_operations(destructive) and not force:
            if not typer.confirm("Continue with potentially destructive changes?"):
                raise typer.Abort()

    # Generate SQL
    sql = ConvertorRegistry.convert_all(statements)

    # Add header
    header = f"""\
-- Migration: {name}
-- Version: {next_version}
-- Generated at: {datetime.now(UTC).isoformat()}
-- Previous: {prev_snapshot.id or "none"}

"""
    full_sql = header + sql

    # Save files
    migrations_dir.mkdir(parents=True, exist_ok=True)

    # Save snapshot and SQL
    folder_path = save_snapshot(
        migrations_dir,
        next_version,
        name,
        current_snapshot.model_dump(mode="json", by_alias=True),
    )
    sql_path = save_migration_sql(folder_path, full_sql)

    # Update journal
    journal.add_entry(version=next_version, tag=name)
    save_journal(journal, migrations_dir)

    typer.echo("")
    typer.echo(f"Created migration: {folder_path.name}/")
    typer.echo(f"  - {sql_path.name}")
    typer.echo("  - snapshot.json")
    typer.echo("")
    typer.echo("Review the migration and run 'derp migrate' to apply.")


def _generate_custom_migration(config: DerpConfig, name: str) -> None:
    """Generate an empty migration for custom SQL."""
    migrations_dir = Path(config.database.migrations_dir)
    journal = load_journal(migrations_dir)

    next_version = get_next_version(journal)

    # Create empty migration
    header = f"""\
-- Migration: {name}
-- Version: {next_version}
-- Generated at: {datetime.now(UTC).isoformat()}
-- Custom migration - add your SQL below

"""

    migrations_dir.mkdir(parents=True, exist_ok=True)

    # Load or create snapshot
    prev_snapshot_data = load_latest_snapshot(migrations_dir, journal)
    if prev_snapshot_data:
        snapshot = prev_snapshot_data
        snapshot["id"] = next_version
        snapshot["prev_id"] = prev_snapshot_data.get("id")
    else:
        snapshot = SchemaSnapshot(id=next_version).model_dump(
            mode="json", by_alias=True
        )

    folder_path = save_snapshot(migrations_dir, next_version, name, snapshot)
    sql_path = save_migration_sql(folder_path, header)

    journal.add_entry(version=next_version, tag=name)
    save_journal(journal, migrations_dir)

    typer.echo(f"Created custom migration: {folder_path.name}/")
    typer.echo(f"Edit {sql_path} to add your SQL statements.")



"""Payments client and typed models."""

from derp.config import PaymentsConfig
from derp.payments.client import PaymentsClient
from derp.payments.exceptions import (
    PaymentsError,
    PaymentsNotConnectedError,
    PaymentsProviderError,
    WebhookSignatureError,
)
from derp.payments.models import (
    Account,
    AccountLink,
    AccountLinkType,
    AccountType,
    Balance,
    CancellationReason,
    CaptureMethod,
    CheckoutSession,
    CheckoutSessionMode,
    Customer,
    PaymentIntent,
    Payout,
    PayoutMethod,
    Refund,
    RefundReason,
    StripeListResult,
    Transfer,
    WebhookEvent,
)

__all__ = [
    "PaymentsClient",
    "PaymentsConfig",
    "PaymentsError",
    "PaymentsProviderError",
    "PaymentsNotConnectedError",
    "WebhookSignatureError",
    "AccountType",
    "AccountLinkType",
    "CaptureMethod",
    "CancellationReason",
    "RefundReason",
    "PayoutMethod",
    "CheckoutSessionMode",
    "Customer",
    "CheckoutSession",
    "StripeListResult",
    "WebhookEvent",
    "Account",
    "AccountLink",
    "Transfer",
    "PaymentIntent",
    "Refund",
    "Payout",
    "Balance",
]



"""Non ORM string-based query builders for Derp."""

from __future__ import annotations

from typing import Any, Self

import asyncpg

from derp.kv.base import KVClient
from derp.orm.query.builder import SelectQuery, _acquire, _WhereShorthandMixin
from derp.orm.query.expressions import Expression
from derp.orm.router import ReplicaRouter

# =============================================================================
# Untyped INSERT
# =============================================================================


class _UntypedInsertBase:
    """Shared INSERT implementation for string-based queries."""

    def __init__(
        self,
        pool: asyncpg.Pool | asyncpg.Connection | None,
        table_name: str,
        *,
        router: ReplicaRouter | None = None,
    ):
        self._pool = pool
        self._table_name = table_name
        self._values: dict[str, Any] = {}
        self._returning_cols: tuple[str, ...] | None = None
        self._router: ReplicaRouter | None = router

    def _build(self) -> tuple[str, list[Any]]:
        columns = list(self._values.keys())
        params = list(self._values.values())
        placeholders = [f"${i + 1}" for i in range(len(params))]

        sql = (
            f"INSERT INTO {self._table_name} ({', '.join(columns)}) "
            f"VALUES ({', '.join(placeholders)})"
        )

        if self._returning_cols:
            sql += f" RETURNING {', '.join(self._returning_cols)}"

        return sql, params


class UntypedInsertQuery(_UntypedInsertBase):
    """INSERT query using string table name — execute() returns None."""

    def returning(self, *columns: str) -> UntypedInsertQueryReturning:
        """Add RETURNING clause. Use '*' for all columns."""
        query = UntypedInsertQueryReturning(
            self._pool, self._table_name, router=self._router
        )
        query._values = self._values
        query._returning_cols = columns
        return query

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> None:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            await conn.execute(sql, *params)
        if self._router is not None:
            self._router.record_write()


class UntypedInsertQueryReturning(_UntypedInsertBase):
    """INSERT with RETURNING — execute() returns dict[str, Any]."""

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> dict[str, Any]:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            row = await conn.fetchrow(sql, *params)
            if row is None:
                raise RuntimeError("INSERT RETURNING returned no rows")
        if self._router is not None:
            self._router.record_write()
        return dict(row)


# =============================================================================
# Untyped UPDATE
# =============================================================================


class _UntypedUpdateBase(_WhereShorthandMixin):
    """Shared UPDATE implementation for string-based queries."""

    def __init__(
        self,
        pool: asyncpg.Pool | asyncpg.Connection | None,
        table_name: str,
        *,
        router: ReplicaRouter | None = None,
    ):
        self._pool = pool
        self._table = table_name
        self._set_values: dict[str, Any] = {}
        self._where_clause: Expression | None = None
        self._returning_cols: tuple[str, ...] | None = None
        self._router: ReplicaRouter | None = router

    def _build(self) -> tuple[str, list[Any]]:
        params: list[Any] = []

        set_parts = []
        for col, val in self._set_values.items():
            params.append(val)
            set_parts.append(f"{col} = ${len(params)}")

        sql = f"UPDATE {self._table} SET {', '.join(set_parts)}"

        if self._where_clause:
            where_sql = self._where_clause.to_sql(params)
            sql += f" WHERE {where_sql}"

        if self._returning_cols:
            sql += f" RETURNING {', '.join(self._returning_cols)}"

        return sql, params


class UntypedUpdateQuery(_UntypedUpdateBase):
    """UPDATE query using string table name — execute() returns None."""

    def set(self, **kwargs: Any) -> Self:
        """Set values to update."""
        self._set_values = kwargs
        return self

    def where(self, cond: Expression) -> Self:
        """Add WHERE clause. Multiple calls combine with AND."""
        if self._where_clause is not None:
            self._where_clause = self._where_clause & cond
        else:
            self._where_clause = cond
        return self

    def returning(self, *columns: str) -> UntypedUpdateQueryReturning:
        """Add RETURNING clause. Use '*' for all columns."""
        query = UntypedUpdateQueryReturning(
            self._pool, self._table, router=self._router
        )
        query._set_values = self._set_values
        query._where_clause = self._where_clause
        query._returning_cols = columns
        return query

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> None:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            await conn.execute(sql, *params)
        if self._router is not None:
            self._router.record_write()


class UntypedUpdateQueryReturning(_UntypedUpdateBase):
    """UPDATE with RETURNING — execute() returns list[dict[str, Any]]."""

    def where(self, cond: Expression) -> Self:
        """Add WHERE clause. Multiple calls combine with AND."""
        if self._where_clause is not None:
            self._where_clause = self._where_clause & cond
        else:
            self._where_clause = cond
        return self

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> list[dict[str, Any]]:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            rows = await conn.fetch(sql, *params)
        if self._router is not None:
            self._router.record_write()
        return [dict(row) for row in rows]


# =============================================================================
# Untyped DELETE
# =============================================================================


class _UntypedDeleteBase(_WhereShorthandMixin):
    """Shared DELETE implementation for string-based queries."""

    def __init__(
        self,
        pool: asyncpg.Pool | asyncpg.Connection | None,
        table_name: str,
        *,
        router: ReplicaRouter | None = None,
    ):
        self._pool = pool
        self._table = table_name
        self._where_clause: Expression | None = None
        self._returning_cols: tuple[str, ...] | None = None
        self._router: ReplicaRouter | None = router

    def _build(self) -> tuple[str, list[Any]]:
        params: list[Any] = []
        sql = f"DELETE FROM {self._table}"

        if self._where_clause:
            where_sql = self._where_clause.to_sql(params)
            sql += f" WHERE {where_sql}"

        if self._returning_cols:
            sql += f" RETURNING {', '.join(self._returning_cols)}"

        return sql, params


class UntypedDeleteQuery(_UntypedDeleteBase):
    """DELETE query using string table name — execute() returns None."""

    def where(self, cond: Expression) -> Self:
        """Add WHERE clause. Multiple calls combine with AND."""
        if self._where_clause is not None:
            self._where_clause = self._where_clause & cond
        else:
            self._where_clause = cond
        return self

    def returning(self, *columns: str) -> UntypedDeleteQueryReturning:
        """Add RETURNING clause. Use '*' for all columns."""
        query = UntypedDeleteQueryReturning(
            self._pool, self._table, router=self._router
        )
        query._where_clause = self._where_clause
        query._returning_cols = columns
        return query

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> None:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            await conn.execute(sql, *params)
        if self._router is not None:
            self._router.record_write()


class UntypedDeleteQueryReturning(_UntypedDeleteBase):
    """DELETE with RETURNING — execute() returns list[dict[str, Any]]."""

    def where(self, cond: Expression) -> Self:
        """Add WHERE clause. Multiple calls combine with AND."""
        if self._where_clause is not None:
            self._where_clause = self._where_clause & cond
        else:
            self._where_clause = cond
        return self

    def build(self) -> tuple[str, list[Any]]:
        return self._build()

    async def execute(self) -> list[dict[str, Any]]:
        if not self._pool:
            raise RuntimeError("No database connection. Call db.connect() first.")
        sql, params = self.build()
        async with _acquire(self._pool) as conn:
            rows = await conn.fetch(sql, *params)
        if self._router is not None:
            self._router.record_write()
        return [dict(row) for row in rows]


# =============================================================================
# TableRef — non ORM entry point
# =============================================================================


class TableRef:
    """Non ORM table reference for string-based queries.

    Example::

        db.table("users").select("*").eq("id", 1).execute()
        db.table("users").insert({"name": "Bob"}).execute()
        db.table("users").update({"name": "Robert"}).eq("id", 1).execute()
        db.table("users").delete().eq("id", 1).execute()
    """

    def __init__(
        self,
        table_name: str,
        pool: asyncpg.Pool | asyncpg.Connection | None,
        *,
        cache_store: KVClient | None = None,
        router: ReplicaRouter | None = None,
    ):
        self._table_name = table_name
        self._pool = pool
        self._cache_store = cache_store
        self._router = router

    def select(self, *columns: str) -> SelectQuery[dict[str, Any]]:
        """Start a SELECT query.

        Args:
            *columns: Column names to select. Use ``"*"`` or omit for all
                columns. A single comma-separated string like
                ``"name, email"`` is also accepted.

        Returns:
            SelectQuery[dict[str, Any]]
        """
        if not columns:
            resolved: tuple[str, ...] = ("*",)
        elif len(columns) == 1 and "," in columns[0]:
            resolved = tuple(c.strip() for c in columns[0].split(","))
        else:
            resolved = columns

        query: SelectQuery[dict[str, Any]] = SelectQuery(
            self._pool,
            resolved,  # type: ignore[arg-type]
            cache_store=self._cache_store,
            router=self._router,
        )
        query._from_table = self._table_name
        return query

    def insert(self, values: dict[str, Any]) -> UntypedInsertQuery:
        """Start an INSERT query.

        Args:
            values: Column-value mapping to insert.

        Returns:
            UntypedInsertQuery
        """
        query = UntypedInsertQuery(self._pool, self._table_name, router=self._router)
        query._values = values
        return query

    def update(self, values: dict[str, Any]) -> UntypedUpdateQuery:
        """Start an UPDATE query.

        Args:
            values: Column-value mapping of SET assignments.

        Returns:
            UntypedUpdateQuery
        """
        query = UntypedUpdateQuery(self._pool, self._table_name, router=self._router)
        query._set_values = values
        return query

    def delete(self) -> UntypedDeleteQuery:
        """Start a DELETE query.

        Returns:
            UntypedDeleteQuery
        """
        return UntypedDeleteQuery(self._pool, self._table_name, router=self._router)



"""Database models for the auth module."""

from __future__ import annotations

import enum
import uuid
from collections.abc import Mapping
from datetime import datetime
from typing import Any, Protocol, runtime_checkable

from pydantic import BaseModel, ConfigDict

from derp.orm import Table
from derp.orm.fields import (
    UUID,
    Boolean,
    Enum,
    Field,
    ForeignKey,
    ForeignKeyAction,
    Text,
    Timestamp,
    Varchar,
)


class AuthProvider(enum.StrEnum):
    """Authentication provider types."""

    EMAIL = "email"
    MAGIC_LINK = "magic_link"
    GOOGLE = "google"
    GITHUB = "github"


@runtime_checkable
class AuthRequest(Protocol):
    """Protocol for objects that carry HTTP headers (e.g. FastAPI Request)."""

    @property
    def headers(self) -> Mapping[str, str]: ...


class UserInfo(BaseModel):
    """Unified user information returned by all auth backends."""

    id: str
    email: str
    first_name: str | None = None
    last_name: str | None = None
    username: str | None = None
    image_url: str | None = None
    role: str
    is_active: bool
    is_superuser: bool
    email_confirmed_at: datetime | None
    last_sign_in_at: datetime | None
    created_at: datetime
    updated_at: datetime
    metadata: dict[str, Any]

    model_config = ConfigDict(frozen=True)


class SessionInfo(BaseModel):
    """Unified session information returned by authenticate."""

    user_id: str
    session_id: str
    role: str
    expires_at: datetime
    metadata: dict[str, Any]
    org_id: str | None = None
    org_role: str | None = None

    model_config = ConfigDict(frozen=True)


class AuthUser(Table, table="users"):
    """User authentication table."""

    id: uuid.UUID = Field(UUID(), primary_key=True, default="gen_random_uuid()")
    email: str = Field(Varchar(255), unique=True, index=True)
    email_confirmed_at: datetime | None = Field(
        Timestamp(with_timezone=True), nullable=True
    )
    encrypted_password: str | None = Field(Text(), nullable=True)
    first_name: str | None = Field(Varchar(255), nullable=True)
    last_name: str | None = Field(Varchar(255), nullable=True)
    username: str | None = Field(Varchar(255), nullable=True)
    image_url: str | None = Field(Text(), nullable=True)
    provider: AuthProvider = Field(Enum(AuthProvider))
    provider_id: str | None = Field(Varchar(255), nullable=True)
    is_active: bool = Field(Boolean(), default=True)
    is_superuser: bool = Field(Boolean(), default=False)
    role: str = Field(Varchar(50), default="default")
    created_at: datetime = Field(Timestamp(with_timezone=True), default="now()")
    updated_at: datetime = Field(Timestamp(with_timezone=True), default="now()")
    last_sign_in_at: datetime | None = Field(
        Timestamp(with_timezone=True), nullable=True
    )


class AuthSession(Table, table="auth_sessions"):
    """Authentication session table with integrated refresh tokens.

    Each row represents a refresh token. Rows sharing the same ``session_id``
    belong to the same logical session (one login event). Token rotation
    inserts a new row and revokes the old one.
    """

    __indexes__ = [("session_id", "revoked")]

    id: uuid.UUID = Field(UUID(), primary_key=True, default="gen_random_uuid()")
    user_id: uuid.UUID = Field(
        UUID(),
        foreign_key=ForeignKey("users.id", on_delete=ForeignKeyAction.CASCADE),
        index=True,
    )
    session_id: uuid.UUID = Field(UUID(), index=True, default="gen_random_uuid()")
    token: str = Field(Varchar(255), unique=True, index=True)
    role: str = Field(Varchar(50), default="default")
    revoked: bool = Field(Boolean(), default=False)
    user_agent: str | None = Field(Text(), nullable=True)
    ip_address: str | None = Field(Varchar(45), nullable=True)  # IPv6 compatible
    org_id: uuid.UUID | None = Field(UUID(), nullable=True)
    not_after: datetime = Field(Timestamp(with_timezone=True))
    created_at: datetime = Field(Timestamp(with_timezone=True), default="now()")


class OrgInfo(BaseModel):
    """Unified organization information returned by all auth backends."""

    id: str
    name: str
    slug: str
    metadata: dict[str, Any]
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(frozen=True)


class OrgMemberInfo(BaseModel):
    """Unified organization membership information."""

    org_id: str
    user_id: str
    role: str
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(frozen=True)


class AuthOrganization(Table, table="organizations"):
    """Organization table for multi-tenancy."""

    id: uuid.UUID = Field(UUID(), primary_key=True, default="gen_random_uuid()")
    name: str = Field(Varchar(255))
    slug: str = Field(Varchar(255), unique=True, index=True)
    metadata: str | None = Field(Text(), nullable=True)
    created_at: datetime = Field(Timestamp(with_timezone=True), default="now()")
    updated_at: datetime = Field(Timestamp(with_timezone=True), default="now()")


class AuthOrgMember(Table, table="org_members"):
    """Organization membership table."""

    __indexes__ = [("org_id", "user_id")]

    id: uuid.UUID = Field(UUID(), primary_key=True, default="gen_random_uuid()")
    org_id: uuid.UUID = Field(
        UUID(),
        foreign_key=ForeignKey("organizations.id", on_delete=ForeignKeyAction.CASCADE),
        index=True,
    )
    user_id: uuid.UUID = Field(
        UUID(),
        foreign_key=ForeignKey("users.id", on_delete=ForeignKeyAction.CASCADE),
        index=True,
    )
    role: str = Field(Varchar(50), default="member")
    created_at: datetime = Field(Timestamp(with_timezone=True), default="now()")
    updated_at: datetime = Field(Timestamp(with_timezone=True), default="now()")



"""Vercel queue client (REST-based)."""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from datetime import timedelta
from typing import Any
from urllib.parse import quote

import httpx

from derp.config import VercelQueueConfig
from derp.queue.base import QueueClient, Schedule, ScheduleType, TaskState, TaskStatus
from derp.queue.exceptions import QueueNotConnectedError, QueueProviderError

VERCEL_API_BASE = "https://api.vercel.com"


class VercelQueueClient(QueueClient):
    """Queue client backed by Vercel Queues (REST API)."""

    supports_result = False
    supports_revoke = False
    supports_delay = True

    def __init__(self, config: VercelQueueConfig):
        self._config = config
        self._client: httpx.AsyncClient | None = None
        self._schedules: list[Schedule] = []

    async def connect(self) -> None:
        if self._client is not None:
            return
        self._client = httpx.AsyncClient(
            base_url=VERCEL_API_BASE,
            headers={
                "Authorization": f"Bearer {self._config.api_token}",
                "Content-Type": "application/json",
            },
        )

    async def disconnect(self) -> None:
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    def _build_params(self) -> dict[str, str]:
        params: dict[str, str] = {}
        if self._config.team_id is not None:
            params["teamId"] = self._config.team_id
        if self._config.project_id is not None:
            params["projectId"] = self._config.project_id
        return params

    async def enqueue(
        self,
        task_name: str,
        payload: dict[str, Any] | None = None,
        *,
        queue: str | None = None,
        delay: timedelta | None = None,
    ) -> str:
        if self._client is None:
            raise QueueNotConnectedError()

        queue_name = queue or self._config.default_queue
        task_id = uuid.uuid4().hex

        body: dict[str, Any] = {
            "task_name": task_name,
            "task_id": task_id,
            "payload": payload or {},
        }
        if delay is not None:
            body["delay_seconds"] = int(delay.total_seconds())

        try:
            resp = await self._client.post(
                f"/v1/queues/{quote(queue_name, safe='')}/messages",
                json=body,
                params=self._build_params(),
            )
            resp.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise QueueProviderError(
                f"Vercel API error: {exc.response.status_code} {exc.response.text}",
                code="vercel_api_error",
            ) from exc
        except Exception as exc:
            raise QueueProviderError(str(exc) or "Failed to enqueue task") from exc

        return task_id

    async def get_status(self, task_id: str) -> TaskStatus:
        """Vercel queues do not expose per-message status."""
        return TaskStatus(
            task_id=task_id,
            state=TaskState.UNKNOWN,
        )

    def register_schedules(self, schedules: Sequence[Schedule]) -> None:
        """Register recurring schedules. Vercel only supports cron expressions."""
        for s in schedules:
            if s.type == ScheduleType.INTERVAL:
                raise QueueProviderError(
                    f"Schedule '{s.name}': Vercel cron only supports "
                    "cron expressions, not intervals.",
                )
        self._schedules = list(schedules)

    def get_schedules(self) -> list[Schedule]:
        """Return the currently registered schedules."""
        return self._schedules

    def generate_vercel_cron_config(self) -> list[dict[str, str]]:
        """Generate the ``crons`` section for vercel.json."""
        return [
            {
                "path": s.path or f"/api/cron/{s.name}",
                "schedule": s.cron or "",
            }
            for s in self._schedules
        ]



"""Statement to SQL convertors.

All convertors auto-register with ConvertorRegistry when imported.
"""

# Import modules to trigger registration (side effect)
from derp.orm.migrations.convertors import column as _column  # noqa: F401
from derp.orm.migrations.convertors import constraint as _constraint  # noqa: F401
from derp.orm.migrations.convertors import enum as _enum  # noqa: F401
from derp.orm.migrations.convertors import index as _index  # noqa: F401
from derp.orm.migrations.convertors import policy as _policy  # noqa: F401
from derp.orm.migrations.convertors import role as _role  # noqa: F401
from derp.orm.migrations.convertors import schema as _schema  # noqa: F401
from derp.orm.migrations.convertors import sequence as _sequence  # noqa: F401
from derp.orm.migrations.convertors import table as _table  # noqa: F401
from derp.orm.migrations.convertors.base import ConvertorRegistry, StatementConvertor

__all__ = [
    "ConvertorRegistry",
    "StatementConvertor",
]



"""Password hashing and validation."""

from __future__ import annotations

import abc
import asyncio
import concurrent.futures
import dataclasses
import re
import secrets

from etils import epy

from derp.config import PasswordConfig

with epy.lazy_imports():
    import argon2
    import argon2.exceptions as argon2_exceptions


class PasswordHasher(abc.ABC):
    """Abstract base class for password hashers."""

    def __init__(self, *, max_workers: int = 8) -> None:
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers, thread_name_prefix="password-hasher"
        )

    @abc.abstractmethod
    def hash(self, password: str) -> str:
        """Hash a password synchronously."""

    @abc.abstractmethod
    def verify(self, password: str, hashed: str) -> bool:
        """Verify a password synchronously."""

    @abc.abstractmethod
    def needs_rehash(self, hashed: str) -> bool:
        """Check if a hash needs to be rehashed (e.g., algorithm upgrade)."""

    async def async_hash(self, password: str) -> str:
        """Hash a password without blocking the event loop."""
        return await asyncio.get_running_loop().run_in_executor(
            self._executor, self.hash, password
        )

    async def async_verify(self, password: str, hashed: str) -> bool:
        """Verify a password without blocking the event loop."""
        return await asyncio.get_running_loop().run_in_executor(
            self._executor, self.verify, password, hashed
        )


class Argon2Hasher(PasswordHasher):
    """Password hasher using Argon2id (recommended)."""

    def __init__(
        self,
        time_cost: int = 3,
        memory_cost: int = 65536,
        parallelism: int = 4,
        *,
        max_workers: int = 8,
    ):
        super().__init__(max_workers=max_workers)
        self._hasher = argon2.PasswordHasher(
            time_cost=time_cost,
            memory_cost=memory_cost,
            parallelism=parallelism,
        )

    def hash(self, password: str) -> str:
        """Hash a password using Argon2id."""
        return self._hasher.hash(password)

    def verify(self, password: str, hashed: str) -> bool:
        """Verify a password against an Argon2 hash."""
        try:
            self._hasher.verify(hashed, password)
            return True
        except (
            argon2_exceptions.VerifyMismatchError,
            argon2_exceptions.InvalidHashError,
        ):
            return False

    def needs_rehash(self, hashed: str) -> bool:
        """Check if the hash needs to be updated with new parameters."""
        return self._hasher.check_needs_rehash(hashed)


@dataclasses.dataclass
class PasswordValidationResult:
    """Result of password validation."""

    valid: bool
    errors: list[str]


def validate_password(
    config: PasswordConfig, password: str
) -> PasswordValidationResult:
    errors: list[str] = []
    if len(password) < config.min_length:
        errors.append(f"Password must be at least {config.min_length} characters")

    if len(password) > config.max_length:
        errors.append(f"Password must be at most {config.max_length} characters")

    if config.require_uppercase and not re.search(r"[A-Z]", password):
        errors.append("Password must contain at least one uppercase letter")

    if config.require_lowercase and not re.search(r"[a-z]", password):
        errors.append("Password must contain at least one lowercase letter")

    if config.require_digit and not re.search(r"\d", password):
        errors.append("Password must contain at least one digit")

    if config.require_special and not re.search(r"[!@#$%^&*(),.?\":{}|<>]", password):
        errors.append("Password must contain at least one special character")

    return PasswordValidationResult(valid=len(errors) == 0, errors=errors)


def generate_secure_token(length: int = 32) -> str:
    """Generate a cryptographically secure random token."""
    return secrets.token_urlsafe(length)



"""Custom exceptions for queue integration."""

from __future__ import annotations


class QueueError(Exception):
    """Base exception for all queue errors."""

    def __init__(self, message: str, code: str | None = None):
        super().__init__(message)
        self.message = message
        self.code = code or "queue_error"


class QueueNotConnectedError(QueueError):
    """Raised when queue client is used before connect()."""

    def __init__(self, message: str = "Queue not connected. Call connect() first."):
        super().__init__(message, code="queue_not_connected")


class QueueProviderError(QueueError):
    """Raised when the queue backend returns an error."""

    def __init__(
        self,
        message: str = "Queue provider request failed",
        code: str | None = None,
    ):
        super().__init__(message, code=code or "queue_provider_error")



"""PostgreSQL field/column type definitions for Derp ORM."""

from __future__ import annotations

import abc
import dataclasses
import datetime
import enum as enum_lib
import re
import uuid as uuid_lib
from collections.abc import Sequence
from decimal import Decimal
from typing import Any

from etils import epy

with epy.lazy_imports():
    from derp.orm.query import expressions


class ForeignKeyAction(enum_lib.StrEnum):
    """Actions for foreign key ON DELETE / ON UPDATE clauses."""

    CASCADE = "CASCADE"
    SET_NULL = "SET NULL"
    SET_DEFAULT = "SET DEFAULT"
    RESTRICT = "RESTRICT"
    NO_ACTION = "NO ACTION"


@dataclasses.dataclass
class FieldType[T](abc.ABC):
    """Base class for PostgreSQL field types, parameterized by Python type."""

    @abc.abstractmethod
    def sql_type(self) -> str:
        """Return the SQL type string for this field."""


# Integer Types


@dataclasses.dataclass
class Serial(FieldType[int]):
    """Auto-incrementing 4-byte integer."""

    def sql_type(self) -> str:
        return "SERIAL"


@dataclasses.dataclass
class BigSerial(FieldType[int]):
    """Auto-incrementing 8-byte integer."""

    def sql_type(self) -> str:
        return "BIGSERIAL"


@dataclasses.dataclass
class SmallInt(FieldType[int]):
    """2-byte signed integer."""

    def sql_type(self) -> str:
        return "SMALLINT"


@dataclasses.dataclass
class Integer(FieldType[int]):
    """4-byte signed integer."""

    def sql_type(self) -> str:
        return "INTEGER"


@dataclasses.dataclass
class BigInt(FieldType[int]):
    """8-byte signed integer."""

    def sql_type(self) -> str:
        return "BIGINT"


# String Types


@dataclasses.dataclass
class Varchar(FieldType[str]):
    """Variable-length string with limit."""

    length: int

    def sql_type(self) -> str:
        return f"VARCHAR({self.length})"


@dataclasses.dataclass
class Char(FieldType[str]):
    """Fixed-length string."""

    length: int

    def sql_type(self) -> str:
        return f"CHAR({self.length})"


@dataclasses.dataclass
class Text(FieldType[str]):
    """Variable unlimited length string."""

    def sql_type(self) -> str:
        return "TEXT"


# Boolean


@dataclasses.dataclass
class Boolean(FieldType[bool]):
    """Boolean type."""

    def sql_type(self) -> str:
        return "BOOLEAN"


# Temporal Types


@dataclasses.dataclass
class Timestamp(FieldType[datetime.datetime]):
    """Timestamp without timezone."""

    with_timezone: bool = False

    def sql_type(self) -> str:
        if self.with_timezone:
            return "TIMESTAMP WITH TIME ZONE"
        return "TIMESTAMP"


@dataclasses.dataclass
class Date(FieldType[datetime.date]):
    """Date type."""

    def sql_type(self) -> str:
        return "DATE"


@dataclasses.dataclass
class Time(FieldType[datetime.time]):
    """Time without timezone."""

    with_timezone: bool = False

    def sql_type(self) -> str:
        if self.with_timezone:
            return "TIME WITH TIME ZONE"
        return "TIME"


@dataclasses.dataclass
class Interval(FieldType[datetime.timedelta]):
    """Time interval."""

    def sql_type(self) -> str:
        return "INTERVAL"


# Numeric Types


@dataclasses.dataclass
class Numeric(FieldType[Decimal]):
    """Exact numeric with precision and scale."""

    precision: int | None = None
    scale: int | None = None

    def sql_type(self) -> str:
        if self.precision is not None and self.scale is not None:
            return f"NUMERIC({self.precision}, {self.scale})"
        elif self.precision is not None:
            return f"NUMERIC({self.precision})"
        return "NUMERIC"


@dataclasses.dataclass
class Real(FieldType[float]):
    """4-byte floating point."""

    def sql_type(self) -> str:
        return "REAL"


@dataclasses.dataclass
class DoublePrecision(FieldType[float]):
    """8-byte floating point."""

    def sql_type(self) -> str:
        return "DOUBLE PRECISION"


# UUID


@dataclasses.dataclass
class UUID(FieldType[uuid_lib.UUID]):
    """UUID type."""

    def sql_type(self) -> str:
        return "UUID"


# Enum


@dataclasses.dataclass
class Enum[E: enum_lib.Enum](FieldType[E]):
    """Enum type."""

    enum: type[E]

    def sql_type(self) -> str:
        return _to_snake_case(self.enum.__name__)


# JSON Types


@dataclasses.dataclass
class JSON(FieldType[Any]):
    """JSON type."""

    def sql_type(self) -> str:
        return "JSON"


@dataclasses.dataclass
class JSONB(FieldType[Any]):
    """Binary JSON type (more efficient for queries)."""

    def sql_type(self) -> str:
        return "JSONB"


# Array Type


@dataclasses.dataclass
class Array[E](FieldType[list[E]]):
    """Array of another type."""

    element_type: FieldType[E]

    def sql_type(self) -> str:
        return f"{self.element_type.sql_type()}[]"


# Vector Types


@dataclasses.dataclass
class Vector(FieldType[list[float]]):
    """Vector type."""

    dim: int

    def sql_type(self) -> str:
        return f"VECTOR({self.dim})"


@dataclasses.dataclass
class TSVector(FieldType[str]):
    """TSVector type."""

    language: str = "english"

    def sql_type(self) -> str:
        return "TSVECTOR"

    def generated(self) -> str:
        return f"to_tsvector('{self.language}')"


# Foreign Key


@dataclasses.dataclass
class ForeignKey:
    """Foreign key reference to another table.column."""

    reference: str | type[Any]
    on_delete: ForeignKeyAction | None = None
    on_update: ForeignKeyAction | None = None

    def to_sql(self) -> str:
        """Generate SQL for foreign key constraint."""
        from derp.orm.table import Table

        if isinstance(self.reference, type) and issubclass(self.reference, Table):
            table: type[Table] = self.reference
            primary_key = table.get_primary_key()
            if primary_key is None:
                raise ValueError(f"Table `{table.__name__}` has no primary key.")
            pk_name, _ = primary_key
            reference = f"{table.get_table_name()}({pk_name})"

        elif not isinstance(self.reference, str):
            raise ValueError(f"Invalid foreign key reference: {self.reference}")
        else:
            reference = self.reference.replace(".", "(") + ")"

        sql = f"REFERENCES {reference}"
        if self.on_delete:
            sql += f" ON DELETE {self.on_delete}"
        if self.on_update:
            sql += f" ON UPDATE {self.on_update}"
        return sql


# Field metadata


@dataclasses.dataclass
class FieldInfo[T]:
    """Metadata for a table field/column, parameterized by Python type."""

    field_type: FieldType[T]
    primary_key: bool = False
    unique: bool = False
    nullable: bool = True
    default: Any = None
    foreign_key: ForeignKey | None = None
    index: bool = False
    # Store references to other FieldInfo for comparison expressions
    _table_name: str | None = dataclasses.field(default=None, repr=False)
    _field_name: str | None = dataclasses.field(default=None, repr=False)

    def is_auto_increment(self) -> bool:
        """Check if this field auto-increments."""
        return isinstance(self.field_type, Serial | BigSerial)

    def __invert__(self) -> Any:
        """Bitwise NOT (~field) — produces ``field = FALSE`` for boolean columns."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.EQ,
            expressions.to_expr(False),
        )

    def __eq__(self, other: Any) -> Any:
        """Equal comparison: field == value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.EQ,
            expressions.to_expr(other),
        )

    def __ne__(self, other: Any) -> Any:
        """Not equal comparison: field != value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.NE,
            expressions.to_expr(other),
        )

    def __lt__(self, other: Any) -> Any:
        """Less than comparison: field < value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.LT,
            expressions.to_expr(other),
        )

    def __le__(self, other: Any) -> Any:
        """Less than or equal comparison: field <= value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.LTE,
            expressions.to_expr(other),
        )

    def __gt__(self, other: Any) -> Any:
        """Greater than comparison: field > value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.GT,
            expressions.to_expr(other),
        )

    def __ge__(self, other: Any) -> Any:
        """Greater than or equal comparison: field >= value."""
        return expressions.BinaryOp(
            self,
            expressions.ComparisonOperator.GTE,
            expressions.to_expr(other),
        )

    def in_(self, values: Sequence[Any] | Any) -> Any:
        """IN clause. Accepts a list of values or a SelectQuery."""
        if hasattr(values, "build"):
            return expressions.InSubquery(self, values, negated=False)
        return expressions.InList(self, tuple(values), negated=False)

    def not_in(self, values: Sequence[Any] | Any) -> Any:
        """NOT IN clause. Accepts a list of values or a SelectQuery."""
        if hasattr(values, "build"):
            return expressions.InSubquery(self, values, negated=True)
        return expressions.InList(self, tuple(values), negated=True)

    def like(self, pattern: str) -> Any:
        """Case-sensitive LIKE pattern matching."""
        return expressions.Like(self, pattern, case_insensitive=False)

    def ilike(self, pattern: str) -> Any:
        """Case-insensitive ILIKE pattern matching."""
        return expressions.Like(self, pattern, case_insensitive=True)

    def is_null(self) -> Any:
        """IS NULL check."""
        return expressions.NullCheck(self, is_null=True)

    def is_not_null(self) -> Any:
        """IS NOT NULL check."""
        return expressions.NullCheck(self, is_null=False)

    def between(self, low: Any, high: Any) -> Any:
        """BETWEEN range check."""
        return expressions.Between(self, low, high)

    def count(self) -> Any:
        """COUNT(column) aggregate."""
        return expressions.AggregateFunc("COUNT", self)

    def sum(self) -> Any:
        """SUM(column) aggregate."""
        return expressions.AggregateFunc("SUM", self)

    def avg(self) -> Any:
        """AVG(column) aggregate."""
        return expressions.AggregateFunc("AVG", self)

    def min(self) -> Any:
        """MIN(column) aggregate."""
        return expressions.AggregateFunc("MIN", self)

    def max(self) -> Any:
        """MAX(column) aggregate."""
        return expressions.AggregateFunc("MAX", self)

    def case(
        self,
        mapping: dict[Any, Any],
        *,
        else_: Any | None = None,
    ) -> Any:
        """Simple CASE expression.

        ``User.c.role.case({"admin": 1, "user": 0}, else_=-1)``
        produces ``CASE users.role WHEN 'admin' THEN 1 ... END``
        """
        return expressions.CaseExpression(self, list(mapping.items()), else_value=else_)


def Field[T](
    t: FieldType[T],
    *,
    primary_key: bool = False,
    unique: bool = False,
    nullable: bool = False,
    default: Any = None,
    foreign_key: ForeignKey | None = None,
    index: bool = False,
) -> T:  # Fool the type checker into validating the type annotation.
    """Create a field definition for a table column.

    Args:
        field_type: The PostgreSQL type for this field
        primary_key: Whether this is the primary key
        unique: Whether values must be unique
        nullable: Whether NULL is allowed (defaults to False for NOT NULL)
        default: Default value or SQL expression (e.g., "now()")
        foreign_key: Foreign key reference
        index: Whether to create an index on this column

    Returns:
        FieldInfo with Pydantic field default (typed as T for compatibility)
    """
    return FieldInfo(  # type: ignore[return-value]
        field_type=t,
        primary_key=primary_key,
        unique=unique,
        nullable=nullable,
        default=default,
        foreign_key=foreign_key,
        index=index,
    )


def _to_snake_case(name: str) -> str:
    """Converts a string (e.g., CamelCase) to snake case."""
    # Use regex to insert an underscore before any uppercase letter that
    # is either preceded by a lowercase letter, or followed by another
    # uppercase letter which is then followed by a lowercase letter. This
    # helps to handle acronyms correctly (e.g., "HTTPHeader" -> "http_header"
    # instead of "h_t_t_p_header").
    pattern = re.compile(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])")
    snake_case_name = pattern.sub("_", name).lower()
    return snake_case_name



"""Row-Level Security (RLS) policy operation convertors."""

from __future__ import annotations

from derp.orm.migrations.convertors.base import (
    ConvertorRegistry,
    StatementConvertor,
    quote_identifier,
    quote_schema_table,
)
from derp.orm.migrations.statements.types import (
    AlterPolicyStatement,
    CreatePolicyStatement,
    DisableRLSStatement,
    DropPolicyStatement,
    EnableRLSStatement,
)


class EnableRLSConvertor(StatementConvertor[EnableRLSStatement]):
    """Convert ENABLE ROW LEVEL SECURITY statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "enable_rls"

    def convert(self, statement: EnableRLSStatement) -> str:
        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        sql = f"ALTER TABLE {table_ref} ENABLE ROW LEVEL SECURITY;"

        if statement.force:
            sql += f"\nALTER TABLE {table_ref} FORCE ROW LEVEL SECURITY;"

        return sql


class DisableRLSConvertor(StatementConvertor[DisableRLSStatement]):
    """Convert DISABLE ROW LEVEL SECURITY statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "disable_rls"

    def convert(self, statement: DisableRLSStatement) -> str:
        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        return f"ALTER TABLE {table_ref} DISABLE ROW LEVEL SECURITY;"


class CreatePolicyConvertor(StatementConvertor[CreatePolicyStatement]):
    """Convert CREATE POLICY statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "create_policy"

    def convert(self, statement: CreatePolicyStatement) -> str:
        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        policy_name = quote_identifier(statement.name)

        parts = [f"CREATE POLICY {policy_name} ON {table_ref}"]

        # AS PERMISSIVE or AS RESTRICTIVE
        if not statement.permissive:
            parts.append("AS RESTRICTIVE")

        # FOR command
        if statement.command and statement.command.upper() != "ALL":
            parts.append(f"FOR {statement.command.upper()}")

        # TO roles
        if statement.roles:
            roles = ", ".join(quote_identifier(r) for r in statement.roles)
            parts.append(f"TO {roles}")

        # USING expression
        if statement.using:
            parts.append(f"USING ({statement.using})")

        # WITH CHECK expression
        if statement.with_check:
            parts.append(f"WITH CHECK ({statement.with_check})")

        return " ".join(parts) + ";"


class DropPolicyConvertor(StatementConvertor[DropPolicyStatement]):
    """Convert DROP POLICY statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "drop_policy"

    def convert(self, statement: DropPolicyStatement) -> str:
        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        policy_name = quote_identifier(statement.name)
        return f"DROP POLICY IF EXISTS {policy_name} ON {table_ref};"


class AlterPolicyConvertor(StatementConvertor[AlterPolicyStatement]):
    """Convert ALTER POLICY statements to SQL."""

    @property
    def statement_type(self) -> str:
        return "alter_policy"

    def convert(self, statement: AlterPolicyStatement) -> str:
        table_ref = quote_schema_table(statement.schema_name, statement.table_name)
        policy_name = quote_identifier(statement.name)

        parts = [f"ALTER POLICY {policy_name} ON {table_ref}"]

        if statement.roles is not None:
            roles = ", ".join(quote_identifier(r) for r in statement.roles)
            parts.append(f"TO {roles}")

        if statement.using is not None:
            parts.append(f"USING ({statement.using})")

        if statement.with_check is not None:
            parts.append(f"WITH CHECK ({statement.with_check})")

        return " ".join(parts) + ";"


# Register convertors
ConvertorRegistry.register(EnableRLSConvertor())
ConvertorRegistry.register(DisableRLSConvertor())
ConvertorRegistry.register(CreatePolicyConvertor())
ConvertorRegistry.register(DropPolicyConvertor())
ConvertorRegistry.register(AlterPolicyConvertor())



"""JSON statement intermediate representation types.

These types represent atomic database operations in a database-agnostic way.
Each statement type maps to one or more SQL statements via convertors.
"""

from __future__ import annotations

from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

# =============================================================================
# Base Statement
# =============================================================================


class JsonStatement(BaseModel):
    """Base class for all JSON statements."""

    type: str

    model_config = ConfigDict(use_enum_values=True)


# =============================================================================
# Column Definition (used in table statements)
# =============================================================================


class ColumnDefinition(BaseModel):
    """Column definition used in CREATE TABLE and ADD COLUMN statements."""

    name: str
    type: str
    primary_key: bool = False
    not_null: bool = True
    unique: bool = False
    default: str | None = None
    generated: str | None = None  # GENERATED ALWAYS AS (expr) STORED
    identity: dict | None = None  # Identity column config
    array_dimensions: int = 0


class UniqueConstraintDefinition(BaseModel):
    """Unique constraint definition."""

    name: str
    columns: list[str]
    nulls_not_distinct: bool = False


class CheckConstraintDefinition(BaseModel):
    """Check constraint definition."""

    name: str
    expression: str


class ForeignKeyDefinition(BaseModel):
    """Foreign key definition."""

    name: str
    columns: list[str]
    references_schema: str = "public"
    references_table: str
    references_columns: list[str]
    on_delete: str | None = None
    on_update: str | None = None
    deferrable: bool = False
    initially_deferred: bool = False


class PrimaryKeyDefinition(BaseModel):
    """Primary key definition."""

    name: str | None = None
    columns: list[str]


# =============================================================================
# Table Statements
# =============================================================================


class CreateTableStatement(JsonStatement):
    """CREATE TABLE statement."""

    type: str = "create_table"
    table_name: str
    schema_name: str = "public"
    columns: list[ColumnDefinition]
    primary_key: PrimaryKeyDefinition | None = None
    unique_constraints: list[UniqueConstraintDefinition] = Field(default_factory=list)
    check_constraints: list[CheckConstraintDefinition] = Field(default_factory=list)
    foreign_keys: list[ForeignKeyDefinition] = Field(default_factory=list)


class DropTableStatement(JsonStatement):
    """DROP TABLE statement."""

    type: str = "drop_table"
    table_name: str
    schema_name: str = "public"
    cascade: bool = False


class RenameTableStatement(JsonStatement):
    """RENAME TABLE statement."""

    type: str = "rename_table"
    from_table: str
    to_table: str
    schema_name: str = "public"


class RecreateTableStatement(JsonStatement):
    """Recreate table (for complex alterations that require recreation)."""

    type: str = "recreate_table"
    table_name: str
    schema_name: str = "public"
    columns: list[ColumnDefinition]
    primary_key: PrimaryKeyDefinition | None = None
    unique_constraints: list[UniqueConstraintDefinition] = Field(default_factory=list)
    check_constraints: list[CheckConstraintDefinition] = Field(default_factory=list)
    foreign_keys: list[ForeignKeyDefinition] = Field(default_factory=list)
    # Data migration: list of (old_column, new_column) mappings
    column_mapping: dict[str, str] = Field(default_factory=dict)


# =============================================================================
# Column Statements
# =============================================================================


class AddColumnStatement(JsonStatement):
    """ALTER TABLE ADD COLUMN statement."""

    type: str = "alter_table_add_column"
    table_name: str
    schema_name: str = "public"
    column: ColumnDefinition


class DropColumnStatement(JsonStatement):
    """ALTER TABLE DROP COLUMN statement."""

    type: str = "alter_table_drop_column"
    table_name: str
    schema_name: str = "public"
    column_name: str
    cascade: bool = False


class RenameColumnStatement(JsonStatement):
    """ALTER TABLE RENAME COLUMN statement."""

    type: str = "alter_table_rename_column"
    table_name: str
    schema_name: str = "public"
    from_column: str
    to_column: str


class AlterColumnTypeStatement(JsonStatement):
    """ALTER TABLE ALTER COLUMN SET DATA TYPE statement."""

    type: str = "alter_table_alter_column_set_type"
    table_name: str
    schema_name: str = "public"
    column_name: str
    old_type: str
    new_type: str
    using: str | None = None  # USING expression for type cast


class AlterColumnNullableStatement(JsonStatement):
    """ALTER TABLE ALTER COLUMN SET/DROP NOT NULL statement."""

    type: str = "alter_table_alter_column_set_nullable"
    table_name: str
    schema_name: str = "public"
    column_name: str
    nullable: bool  # True = DROP NOT NULL, False = SET NOT NULL


class AlterColumnDefaultStatement(JsonStatement):
    """ALTER TABLE ALTER COLUMN SET/DROP DEFAULT statement."""

    type: str = "alter_table_alter_column_set_default"
    table_name: str
    schema_name: str = "public"
    column_name: str
    default: str | None  # None = DROP DEFAULT


# =============================================================================
# Foreign Key Statements
# =============================================================================


class CreateForeignKeyStatement(JsonStatement):
    """ALTER TABLE ADD CONSTRAINT ... FOREIGN KEY statement."""

    type: str = "create_foreign_key"
    name: str
    table_name: str
    schema_name: str = "public"
    columns: list[str]
    references_schema: str = "public"
    references_table: str
    references_columns: list[str]
    on_delete: str | None = None
    on_update: str | None = None
    deferrable: bool = False
    initially_deferred: bool = False


class DropForeignKeyStatement(JsonStatement):
    """ALTER TABLE DROP CONSTRAINT (foreign key) statement."""

    type: str = "drop_foreign_key"
    name: str
    table_name: str
    schema_name: str = "public"


# =============================================================================
# Unique Constraint Statements
# =============================================================================


class CreateUniqueConstraintStatement(JsonStatement):
    """ALTER TABLE ADD CONSTRAINT ... UNIQUE statement."""

    type: str = "create_unique_constraint"
    name: str
    table_name: str
    schema_name: str = "public"
    columns: list[str]
    nulls_not_distinct: bool = False


class DropUniqueConstraintStatement(JsonStatement):
    """ALTER TABLE DROP CONSTRAINT (unique) statement."""

    type: str = "drop_unique_constraint"
    name: str
    table_name: str
    schema_name: str = "public"


# =============================================================================
# Check Constraint Statements
# =============================================================================


class CreateCheckConstraintStatement(JsonStatement):
    """ALTER TABLE ADD CONSTRAINT ... CHECK statement."""

    type: str = "create_check_constraint"
    name: str
    table_name: str
    schema_name: str = "public"
    expression: str


class DropCheckConstraintStatement(JsonStatement):
    """ALTER TABLE DROP CONSTRAINT (check) statement."""

    type: str = "drop_check_constraint"
    name: str
    table_name: str
    schema_name: str = "public"


# =============================================================================
# Primary Key Statements
# =============================================================================


class CreatePrimaryKeyStatement(JsonStatement):
    """ALTER TABLE ADD PRIMARY KEY statement."""

    type: str = "create_pk"
    name: str | None = None
    table_name: str
    schema_name: str = "public"
    columns: list[str]


class DropPrimaryKeyStatement(JsonStatement):
    """ALTER TABLE DROP CONSTRAINT (primary key) statement."""

    type: str = "drop_pk"
    name: str
    table_name: str
    schema_name: str = "public"


# =============================================================================
# Index Statements
# =============================================================================


class CreateIndexStatement(JsonStatement):
    """CREATE INDEX statement."""

    type: str = "create_index"
    name: str
    table_name: str
    schema_name: str = "public"
    columns: list[str]
    unique: bool = False
    where: str | None = None  # Partial index condition
    method: str = "btree"  # btree, hash, gin, gist, etc.
    concurrently: bool = False
    nulls_not_distinct: bool = False
    include: list[str] = Field(default_factory=list)  # INCLUDE columns


class DropIndexStatement(JsonStatement):
    """DROP INDEX statement."""

    type: str = "drop_index"
    name: str
    schema_name: str = "public"
    concurrently: bool = False


# =============================================================================
# Enum Statements
# =============================================================================


class CreateEnumStatement(JsonStatement):
    """CREATE TYPE ... AS ENUM statement."""

    type: str = "create_enum"
    name: str
    schema_name: str = "public"
    values: list[str]


class DropEnumStatement(JsonStatement):
    """DROP TYPE (enum) statement."""

    type: str = "drop_enum"
    name: str
    schema_name: str = "public"
    cascade: bool = False


class AlterEnumAddValueStatement(JsonStatement):
    """ALTER TYPE ... ADD VALUE statement."""

    type: str = "alter_enum_add_value"
    name: str
    schema_name: str = "public"
    value: str
    before: str | None = None  # Add before this value
    after: str | None = None  # Add after this value


class AlterEnumRenameValueStatement(JsonStatement):
    """ALTER TYPE ... RENAME VALUE statement (PostgreSQL 10+)."""

    type: str = "alter_enum_rename_value"
    name: str
    schema_name: str = "public"
    old_value: str
    new_value: str


# =============================================================================
# Sequence Statements
# =============================================================================


class CreateSequenceStatement(JsonStatement):
    """CREATE SEQUENCE statement."""

    type: str = "create_sequence"
    name: str
    schema_name: str = "public"
    start: int = 1
    increment: int = 1
    min_value: int | None = None
    max_value: int | None = None
    cache: int = 1
    cycle: bool = False
    owned_by: str | None = None  # "table.column"


class DropSequenceStatement(JsonStatement):
    """DROP SEQUENCE statement."""

    type: str = "drop_sequence"
    name: str
    schema_name: str = "public"
    cascade: bool = False


class AlterSequenceStatement(JsonStatement):
    """ALTER SEQUENCE statement."""

    type: str = "alter_sequence"
    name: str
    schema_name: str = "public"
    restart: int | None = None
    increment: int | None = None
    min_value: int | None = None
    max_value: int | None = None
    cache: int | None = None
    cycle: bool | None = None
    owned_by: str | None = None


# =============================================================================
# Schema Statements
# =============================================================================


class CreateSchemaStatement(JsonStatement):
    """CREATE SCHEMA statement."""

    type: str = "create_schema"
    name: str
    authorization: str | None = None


class DropSchemaStatement(JsonStatement):
    """DROP SCHEMA statement."""

    type: str = "drop_schema"
    name: str
    cascade: bool = False


# =============================================================================
# Row-Level Security Statements
# =============================================================================


class EnableRLSStatement(JsonStatement):
    """ALTER TABLE ENABLE ROW LEVEL SECURITY statement."""

    type: str = "enable_rls"
    table_name: str
    schema_name: str = "public"
    force: bool = False  # FORCE vs not


class DisableRLSStatement(JsonStatement):
    """ALTER TABLE DISABLE ROW LEVEL SECURITY statement."""

    type: str = "disable_rls"
    table_name: str
    schema_name: str = "public"


class CreatePolicyStatement(JsonStatement):
    """CREATE POLICY statement."""

    type: str = "create_policy"
    name: str
    table_name: str
    schema_name: str = "public"
    command: str = "ALL"  # ALL, SELECT, INSERT, UPDATE, DELETE
    permissive: bool = True  # PERMISSIVE vs RESTRICTIVE
    roles: list[str] = Field(default_factory=lambda: ["public"])
    using: str | None = None  # USING expression
    with_check: str | None = None  # WITH CHECK expression


class DropPolicyStatement(JsonStatement):
    """DROP POLICY statement."""

    type: str = "drop_policy"
    name: str
    table_name: str
    schema_name: str = "public"


class AlterPolicyStatement(JsonStatement):
    """ALTER POLICY statement."""

    type: str = "alter_policy"
    name: str
    table_name: str
    schema_name: str = "public"
    roles: list[str] | None = None
    using: str | None = None
    with_check: str | None = None


# =============================================================================
# Role Statements
# =============================================================================


class CreateRoleStatement(JsonStatement):
    """CREATE ROLE statement."""

    type: str = "create_role"
    name: str
    superuser: bool = False
    create_db: bool = False
    create_role: bool = False
    inherit: bool = True
    login: bool = False
    replication: bool = False
    bypass_rls: bool = False
    connection_limit: int = -1
    password: str | None = None
    valid_until: str | None = None
    in_roles: list[str] = Field(default_factory=list)


class DropRoleStatement(JsonStatement):
    """DROP ROLE statement."""

    type: str = "drop_role"
    name: str


class GrantStatement(JsonStatement):
    """GRANT statement."""

    type: str = "grant"
    privileges: list[str]  # SELECT, INSERT, UPDATE, DELETE, ALL, etc.
    object_type: str  # TABLE, SCHEMA, SEQUENCE, FUNCTION
    object_schema: str = "public"
    object_name: str
    grantee: str  # Role name
    with_grant_option: bool = False


class RevokeStatement(JsonStatement):
    """REVOKE statement."""

    type: str = "revoke"
    privileges: list[str]
    object_type: str
    object_schema: str = "public"
    object_name: str
    grantee: str
    cascade: bool = False


# =============================================================================
# Statement Union Type
# =============================================================================


Statement = Annotated[
    (
        # Table
        CreateTableStatement
        | DropTableStatement
        | RenameTableStatement
        | RecreateTableStatement
        # Column
        | AddColumnStatement
        | DropColumnStatement
        | RenameColumnStatement
        | AlterColumnTypeStatement
        | AlterColumnNullableStatement
        | AlterColumnDefaultStatement
        # Foreign Key
        | CreateForeignKeyStatement
        | DropForeignKeyStatement
        # Unique Constraint
        | CreateUniqueConstraintStatement
        | DropUniqueConstraintStatement
        # Check Constraint
        | CreateCheckConstraintStatement
        | DropCheckConstraintStatement
        # Primary Key
        | CreatePrimaryKeyStatement
        | DropPrimaryKeyStatement
        # Index
        | CreateIndexStatement
        | DropIndexStatement
        # Enum
        | CreateEnumStatement
        | DropEnumStatement
        | AlterEnumAddValueStatement
        | AlterEnumRenameValueStatement
        # Sequence
        | CreateSequenceStatement
        | DropSequenceStatement
        | AlterSequenceStatement
        # Schema
        | CreateSchemaStatement
        | DropSchemaStatement
        # RLS
        | EnableRLSStatement
        | DisableRLSStatement
        | CreatePolicyStatement
        | DropPolicyStatement
        | AlterPolicyStatement
        # Role
        | CreateRoleStatement
        | DropRoleStatement
        | GrantStatement
        | RevokeStatement
    ),
    Field(discriminator="type"),
]



"""OAuth provider implementations."""

from __future__ import annotations

from derp.auth.providers.base import BaseOAuthProvider, OAuthUserInfo
from derp.auth.providers.github import GitHubProvider
from derp.auth.providers.google import GoogleProvider

__all__ = [
    "BaseOAuthProvider",
    "GitHubProvider",
    "GoogleProvider",
    "OAuthUserInfo",
]



"""Derp Auth - Authentication library for FastAPI applications."""

from __future__ import annotations

from derp.auth.base import BaseAuthClient
from derp.auth.clerk_client import ClerkAuthClient
from derp.auth.email import EmailClient
from derp.auth.exceptions import (
    AuthError,
    ConfirmationTokenInvalidError,
    EmailNotConfirmedError,
    EmailSendError,
    InvalidCredentialsError,
    InvalidTokenError,
    MagicLinkExpiredError,
    NotOrgMemberError,
    OAuthError,
    OAuthProviderError,
    OAuthStateError,
    OrgAlreadyExistsError,
    OrgLastOwnerError,
    OrgMemberExistsError,
    OrgMemberNotFoundError,
    OrgNotFoundError,
    PasswordValidationError,
    RecoveryTokenInvalidError,
    RefreshTokenReusedError,
    RefreshTokenRevokedError,
    SessionExpiredError,
    SessionNotFoundError,
    SignupDisabledError,
    TokenExpiredError,
    UserAlreadyExistsError,
    UserNotActiveError,
    UserNotFoundError,
)
from derp.auth.jwt import TokenPair, TokenPayload
from derp.auth.models import (
    AuthOrganization,
    AuthOrgMember,
    AuthProvider,
    AuthRequest,
    AuthSession,
    AuthUser,
    OrgInfo,
    OrgMemberInfo,
    SessionInfo,
    UserInfo,
)
from derp.auth.native_client import NativeAuthClient
from derp.auth.password import (
    Argon2Hasher,
    PasswordHasher,
    PasswordValidationResult,
    generate_secure_token,
)
from derp.auth.providers import (
    BaseOAuthProvider,
    GitHubProvider,
    GoogleProvider,
    OAuthUserInfo,
)
from derp.config import (
    AuthConfig,
    ClerkConfig,
    EmailConfig,
    GitHubOAuthConfig,
    GoogleOAuthConfig,
    JWTConfig,
    NativeAuthConfig,
    PasswordConfig,
)

__all__ = [
    # Config
    "AuthConfig",
    "ClerkConfig",
    "EmailConfig",
    "GitHubOAuthConfig",
    "GoogleOAuthConfig",
    "JWTConfig",
    "NativeAuthConfig",
    "PasswordConfig",
    # Exceptions
    "AuthError",
    "ConfirmationTokenInvalidError",
    "EmailNotConfirmedError",
    "EmailSendError",
    "InvalidCredentialsError",
    "InvalidTokenError",
    "MagicLinkExpiredError",
    "NotOrgMemberError",
    "OAuthError",
    "OAuthProviderError",
    "OAuthStateError",
    "OrgAlreadyExistsError",
    "OrgLastOwnerError",
    "OrgMemberExistsError",
    "OrgMemberNotFoundError",
    "OrgNotFoundError",
    "PasswordValidationError",
    "RecoveryTokenInvalidError",
    "RefreshTokenRevokedError",
    "RefreshTokenReusedError",
    "SessionExpiredError",
    "SessionNotFoundError",
    "SignupDisabledError",
    "TokenExpiredError",
    "UserAlreadyExistsError",
    "UserNotActiveError",
    "UserNotFoundError",
    # JWT
    "JWTManager",
    "TokenPair",
    "TokenPayload",
    # Models
    "AuthOrgMember",
    "AuthOrganization",
    "AuthProvider",
    "AuthRequest",
    "AuthSession",
    "AuthUser",
    "OrgInfo",
    "OrgMemberInfo",
    "SessionInfo",
    "UserInfo",
    # Password
    "Argon2Hasher",
    "BcryptHasher",
    "CompositeHasher",
    "PasswordHasher",
    "PasswordValidationResult",
    "PasswordValidator",
    "create_default_hasher",
    "generate_secure_token",
    # OAuth Providers
    "BaseOAuthProvider",
    "GitHubProvider",
    "GoogleProvider",
    "OAuthUserInfo",
    # Email
    "EmailClient",
    # Client
    "BaseAuthClient",
    "ClerkAuthClient",
    "NativeAuthClient",
]



"""Storage wrapper for S3-compatible object storage.

Example usage:

    from storage import Storage

    async with Storage(
        bucket_name="my-bucket",
        endpoint_url="https://s3.amazonaws.com",
        aws_access_key_id="key",
        aws_secret_access_key="secret",
    ) as storage:
        # Upload a file
        await storage.upload_file("path/to/local/file.txt", "remote/file.txt")

        # Fetch a file
        content = await storage.fetch_file("remote/file.txt")
"""

from derp.config import StorageConfig

from .client import StorageClient

__all__ = ["StorageClient", "StorageConfig"]
